"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements sensors for CWH platforms.
"""
import typing

import numpy as np
from corl.libraries.plugin_library import PluginLibrary
from corl.libraries.units import corl_get_ureg
from corl.simulators.base_parts import BasePlatformPartValidator, BaseSensor

import safe_autonomy_sims.platforms.cwh.cwh_properties as cwh_props
from safe_autonomy_sims.platforms.cwh.cwh_available_platforms import CWHAvailablePlatformTypes
from safe_autonomy_sims.simulators.cwh_simulator import CWHSimulator
from safe_autonomy_sims.simulators.inspection_simulator import InspectionSimulator


class CWHSensor(BaseSensor):
    """
    Interface for a basic sensor of the CWH platform.
    """

    def _calculate_measurement(self, state):
        """
        Get measurements from the sensor.

        Raises
        ------
        NotImplementedError
            If the method has not been implemented.
        """
        raise NotImplementedError


class PositionSensor(CWHSensor):
    """
    Implementation of a sensor designed to give the position at any time.
    """

    def __init__(self, parent_platform, config, property_class=cwh_props.PositionProp):
        super().__init__(property_class=property_class, parent_platform=parent_platform, config=config)

    def _calculate_measurement(self, state):
        """
        Calculate the measurement - position.

        Returns
        -------
        Quantity
            Position of spacecraft.
        """
        position = np.array(self.parent_platform.position, dtype=np.float32)
        return corl_get_ureg().Quantity(position, "m")


class RelativePositionSensorValidator(BasePlatformPartValidator):
    """
    A configuration validator for RelativePositionSensor

    Attributes
    ----------
    entity_name: str
        The name of the entity the position of which is to be returned.
    """
    entity_name: str = "chief"


class RelativePositionSensor(CWHSensor):
    """
    Implementation of a sensor designed to give the position at any time.
    """

    def __init__(self, parent_platform, config, property_class=cwh_props.RelativePositionProp):
        super().__init__(property_class=property_class, parent_platform=parent_platform, config=config)

    @staticmethod
    def get_validator() -> typing.Type[BasePlatformPartValidator]:
        """
        return the validator that will be used on the configuration
        of this part
        """
        return RelativePositionSensorValidator

    def _calculate_measurement(self, state):
        """
        Calculate the measurement - position.

        Returns
        -------
        list of floats
            Position of spacecraft.
        """
        relative_position = None
        # handle initialization case
        if self.config.entity_name not in state.sim_entities:
            # raise error if not initialization
            if state.sim_time != 0.0:
                raise ValueError(f"{self.config.entity_name} not found in simulator state!")
            relative_position = np.array([0.0, 0.0, 0.0])
        else:
            relative_position = self.parent_platform.entity_relative_position(state.sim_entities[self.config.entity_name])

        return corl_get_ureg().Quantity(np.array(relative_position, dtype=np.float32), "m")


class VelocitySensor(CWHSensor):
    """
    Implementation of a sensor to give velocity at any time.
    """

    def __init__(self, parent_platform, config, property_class=cwh_props.VelocityProp):
        super().__init__(property_class=property_class, parent_platform=parent_platform, config=config)

    # state - tuple
    def _calculate_measurement(self, state):
        """
        Calculate the measurement - velocity.

        Returns
        -------
        list of floats
            Velocity of spacecraft.
        """
        velocity = np.array(self.parent_platform.velocity, dtype=np.float32)
        return corl_get_ureg().Quantity(velocity, "meter/second")


class RelativeVelocitySensorValidator(BasePlatformPartValidator):
    """
    A configuration validator for RelativeVelocitySensor

    Attributes
    ----------
    entity_name: str
        The name of the entity the velocity of which is to be returned.
    """
    entity_name: str = "chief"


class RelativeVelocitySensor(CWHSensor):
    """
    Implementation of a sensor designed to give the relative velocity at any
    time.
    """

    def __init__(self, parent_platform, config, property_class=cwh_props.RelativeVelocityProp):
        super().__init__(property_class=property_class, parent_platform=parent_platform, config=config)

    @staticmethod
    def get_validator() -> typing.Type[BasePlatformPartValidator]:
        """
        return the validator that will be used on the configuration
        of this part
        """
        return RelativeVelocitySensorValidator

    def _calculate_measurement(self, state):
        """
        Calculate the measurement - position.

        Returns
        -------
        list of floats
            Position of spacecraft.
        """
        relative_velocity = None
        # handle initialization case
        if self.config.entity_name not in state.sim_entities:
            # raise error if not initialization
            if state.sim_time != 0.0:
                raise ValueError(f"{self.config.entity_name} not found in simulator state!")
            relative_velocity = np.array([0.0, 0.0, 0.0])
        else:
            relative_velocity = self.parent_platform.entity_relative_velocity(state.sim_entities[self.config.entity_name])

        return corl_get_ureg().Quantity(np.array(relative_velocity, dtype=np.float32), "m/s")


class InspectedPointsSensorValidator(BasePlatformPartValidator):
    """
    A configuration validator for InspectedPointsSensor

    Attributes
    ----------
    inspector_entity_name: str
        The name of the entity performing inspection.
    """
    inspector_entity_name: str = ""


class InspectedPointsSensor(CWHSensor):
    """
    Implementation of a sensor to give number of points at any time.
    """

    def __init__(self, parent_platform, config, property_class=cwh_props.InspectedPointProp):
        super().__init__(property_class=property_class, parent_platform=parent_platform, config=config)

    @staticmethod
    def get_validator() -> typing.Type[BasePlatformPartValidator]:
        """
        return the validator that will be used on the configuration
        of this part
        """
        return InspectedPointsSensorValidator

    def _calculate_measurement(self, state):
        """
        Calculate the measurement - num_inspected_points.

        Returns
        -------
        int
            Number of inspected points.
        """
        num_points_inspected = 0
        # handle initialization case
        if self.config.inspector_entity_name not in state.sim_entities:
            # raise error if not initialization
            if state.sim_time != 0.0:
                raise ValueError(f"{self.config.inspector_entity_name} not found in simulator state!")
        else:
            # get inspector entity object
            inspector_entity = state.sim_entities[self.config.inspector_entity_name]

            # count total number of points inspected by
            for points in state.inspection_points_map.values():
                num_points_inspected += points.get_num_points_inspected(inspector_entity=inspector_entity)

        return corl_get_ureg().Quantity(np.array([num_points_inspected], dtype=np.float32), "dimensionless")


class SunAngleSensor(CWHSensor):
    """
    Implementation of a sensor to give the sun angle
    """

    def __init__(self, parent_platform, config, property_class=cwh_props.SunAngleProp):
        super().__init__(property_class=property_class, parent_platform=parent_platform, config=config)

    def _calculate_measurement(self, state):
        """
        Calculate the measurement - sun angle.

        Returns
        -------
        float
            sun angle
        """
        sun_angle = np.array([state.sun_angle], dtype=np.float32)
        return corl_get_ureg().Quantity(sun_angle, "radians")


class SunVectorSensor(CWHSensor):
    """
    Implementation of a sensor to give the sun unit vector
    """

    def __init__(self, parent_platform, config, property_class=cwh_props.SunVectorProp):
        super().__init__(property_class=property_class, parent_platform=parent_platform, config=config)

    def _calculate_measurement(self, state):
        """
        Calculate the measurement - sun angle.

        Returns
        -------
        float
            sun angle
        """
        sun_position = np.array([np.cos(state.sun_angle), -np.sin(state.sun_angle), 0], dtype=np.float32)
        sun_position = corl_get_ureg().Quantity(sun_position, "m")
        return sun_position


class UninspectedPointsSensorValidator(BasePlatformPartValidator):
    """
    A configuration validator for InspectedPointsSensor

    Attributes
    ----------
    inspector_entity_name: str
        The name of the entity performing inspection.
    inspection_entity_name: str
        The name of the entity under inspection.
    """
    inspector_entity_name: str = ""
    inspection_entity_name: str = ""


class UninspectedPointsSensor(CWHSensor):
    """
    Implementation of a sensor to give direction from the origin to the location
    of a cluster of uninspected points.
    """

    def __init__(self, parent_platform, config, property_class=cwh_props.UninspectedPointProp):
        super().__init__(property_class=property_class, parent_platform=parent_platform, config=config)

    @staticmethod
    def get_validator() -> typing.Type[BasePlatformPartValidator]:
        """
        return the validator that will be used on the configuration
        of this part
        """
        return UninspectedPointsSensorValidator

    def _calculate_measurement(self, state):
        """
        Calculate the measurement - cluster_location.

        Returns
        -------
        np.ndarray
            Cluster location of the uninspected points.
        """
        # handle initialization case
        if (
            self.config.inspection_entity_name not in state.inspection_points_map
            or self.config.inspector_entity_name not in state.sim_entities
        ):
            # raise error if not initialization
            if state.sim_time != 0.0:
                raise ValueError(
                    f"{self.config.inspector_entity_name} not found in simulator state \
                        or {self.config.inspection_entity_name} is not an inspectable entity!"
                )
            return corl_get_ureg().Quantity([0., 0., 0.], "meters")

        # get entities
        inspector_entity = state.sim_entities[self.config.inspector_entity_name]

        # get inspection points of object under inspection
        inspector_position = inspector_entity.position
        inspection_points = state.inspection_points_map[self.config.inspection_entity_name]
        cluster_position = inspection_points.kmeans_find_nearest_cluster(inspector_position)
        cluster_position = np.array(cluster_position, dtype=np.float32)
        return corl_get_ureg().Quantity(cluster_position, "meters")


# entity position sensors
class EntitySensorValidator(BasePlatformPartValidator):
    """
    A configuration validator for the various EntitySensors

    Attributes
    ----------
    entity_name : str
        name of the entity to sense
    """
    entity_name: str


class EntityPositionSensor(CWHSensor):
    """
    Implementation of a sensor designed to give the position at any time.
    """

    def __init__(self, parent_platform, config, property_class=cwh_props.PositionProp):
        super().__init__(property_class=property_class, parent_platform=parent_platform, config=config)

    @staticmethod
    def get_validator() -> typing.Type[EntitySensorValidator]:
        """
        return the validator that will be used on the configuration
        of this part
        """
        return EntitySensorValidator

    def _calculate_measurement(self, state):
        """
        Calculate the measurement - position.

        Returns
        -------
        list of floats
            Position of spacecraft.
        """
        return corl_get_ureg().Quantity(state.sim_entities[self.config.entity_name].position, "meters")


class OriginPositionSensor(CWHSensor):
    """
    Implementation of a sensor designed to give the position at any time.
    """

    def __init__(self, parent_platform, config, property_class=cwh_props.PositionProp):
        super().__init__(property_class=property_class, parent_platform=parent_platform, config=config)

    def _calculate_measurement(self, state):
        """
        Calculate the measurement - position.

        Returns
        -------
        list of floats
            Position of spacecraft.
        """
        return corl_get_ureg().Quantity(np.array([0, 0, 0]), "meters")


class EntityVelocitySensor(CWHSensor):
    """
    Implementation of a sensor designed to give the position at any time.
    """

    def __init__(self, parent_platform, config, property_class=cwh_props.VelocityProp):
        super().__init__(property_class=property_class, parent_platform=parent_platform, config=config)

    @staticmethod
    def get_validator() -> typing.Type[EntitySensorValidator]:
        """
        return the validator that will be used on the configuration
        of this part
        """
        return EntitySensorValidator

    def _calculate_measurement(self, state):
        """
        Calculate the measurement - position.

        Returns
        -------
        list of floats
            Position of spacecraft.
        """
        return corl_get_ureg().Quantity(state.sim_entities[self.config.entity_name].velocity, "meter/second")


class BoolArraySensorValidator(BasePlatformPartValidator):
    """
    A configuration validator for BoolArraySensor.

    Attributes
    ----------
    inspection_entity_name: str
        The name of the entity under inspection.
    """
    inspection_entity_name: str = ""


class BoolArraySensor(CWHSensor):
    """
    Implementation of a sensor to give boolean array for all inspected/uninspected points
    """

    def __init__(self, parent_platform, config, property_class=cwh_props.BoolArrayProp):
        super().__init__(property_class=property_class, parent_platform=parent_platform, config=config)

    @staticmethod
    def get_validator() -> typing.Type[BasePlatformPartValidator]:
        """
        return the validator that will be used on the configuration
        of this part
        """
        return BoolArraySensorValidator

    def _calculate_measurement(self, state):
        """
        Calculate the measurement - bool_array.

        Returns
        -------
        np.ndarray
            Bool array describing inspected/uninspected points.
        """
        bool_array = None
        # handle initialization case
        if self.config.inspection_entity_name not in state.inspection_points_map:
            # raise error if not initialization
            if state.sim_time != 0.0:
                raise ValueError(f"{self.config.inspection_entity_name} is not an inspectable entity!")
            bool_array = np.array([float(False)])
        else:
            inspection_points = state.inspection_points_map[self.config.inspection_entity_name]
            bool_array = np.array([float(bool(a)) for a in inspection_points.points_inspected_dict.values()])
            if len(bool_array) == 99:  # TODO: Remove hardcoded value
                bool_array = np.concatenate((bool_array, np.zeros(1)))

        return corl_get_ureg().Quantity(bool_array, "dimensionless")


class PriorityVectorSensor(CWHSensor):
    """
    Implementation of a sensor to give the inspected points priority vector
    """

    def __init__(self, parent_platform, config, property_class=cwh_props.PriorityVectorProp):
        super().__init__(property_class=property_class, parent_platform=parent_platform, config=config)

    def _calculate_measurement(self, state):
        """
        Calculate the measurement - priority vector.

        Returns
        -------
        np.ndarray
            priority vector
        """
        priority_vector = np.array(state.priority_vector, dtype=np.float32)
        return corl_get_ureg().Quantity(priority_vector, "m")


class InspectedPointsScoreSensorValidator(BasePlatformPartValidator):
    """
    A configuration validator for InspectedPointsScoreSensor

    Attributes
    ----------
    inspection_entity_name: str
        The name of the entity under inspection.
    """
    inspector_entity_name: str = ""


class InspectedPointsScoreSensor(CWHSensor):
    """
    Implementation of a sensor to give the inspected points score
    """

    def __init__(self, parent_platform, config, property_class=cwh_props.PointsScoreProp):
        super().__init__(property_class=property_class, parent_platform=parent_platform, config=config)

    @staticmethod
    def get_validator() -> typing.Type[BasePlatformPartValidator]:
        """
        return the validator that will be used on the configuration
        of this part
        """
        return InspectedPointsScoreSensorValidator

    def _calculate_measurement(self, state):
        """
        Calculate the measurement - inspected points score.

        Returns
        -------
        float
            inspected points score.
        """
        # handle initialization case
        if self.config.inspector_entity_name not in state.sim_entities:
            # raise error if not initialization
            if state.sim_time != 0.0:
                raise ValueError(f"{self.config.inspector_entity_name} not found in simulator state!")
            return corl_get_ureg().Quantity(np.array([0.]), "dimensionless")

        # get inspector entity object
        inspector_entity = state.sim_entities[self.config.inspector_entity_name]

        # count total number of points inspected by
        weight = 0.
        for points in state.inspection_points_map.values():
            weight += points.get_total_weight_inspected(inspector_entity=inspector_entity)

        return corl_get_ureg().Quantity(np.array([weight], dtype=np.float32), "dimensionless")


class OrbitStabilitySensor(CWHSensor):
    """
    Implementation of a sensor to give 2nx + v_y, the quantity that determines
    stability of the un-controlled motion
    """

    def __init__(self, parent_platform, config, property_class=cwh_props.OrbitStabilityProp):
        super().__init__(property_class=property_class, parent_platform=parent_platform, config=config)

    def _calculate_measurement(self, state):
        """
        Calculate the measurement - cluster_location.

        Returns
        -------
        np.ndarray
            Cluster location of the uninspected points.
        """
        pos = self.parent_platform.position
        vel = self.parent_platform.velocity

        n = self.parent_platform._platform.dynamics.n  # pylint: disable=protected-access

        orbit_stability = 2 * pos[0] * n + vel[1]

        return corl_get_ureg().Quantity(np.array([orbit_stability], dtype=np.float32), "dimensionless")


for sim in [CWHSimulator, InspectionSimulator]:
    for sensor, sensor_name in zip(
        [
            CWHSensor,
            PositionSensor,
            VelocitySensor,
            RelativePositionSensor,
            RelativeVelocitySensor,
            InspectedPointsSensor,
            SunAngleSensor,
            SunVectorSensor,
            UninspectedPointsSensor,
            BoolArraySensor,
            EntityPositionSensor,
            EntityVelocitySensor,
            OriginPositionSensor,
            PriorityVectorSensor,
            InspectedPointsScoreSensor,
            OrbitStabilitySensor,
        ],
        [
            "Sensor_Generic",
            "Sensor_Position",
            "Sensor_Velocity",
            "Sensor_RelativePosition",
            "Sensor_RelativeVelocity",
            "Sensor_InspectedPoints",
            "Sensor_SunAngle",
            "Sensor_SunVector",
            "Sensor_UninspectedPoints",
            "Sensor_BoolArray",
            "Sensor_EntityPosition",
            "Sensor_EntityVelocity",
            "Sensor_OriginPosition",
            "Sensor_PriorityVector",
            "Sensor_InspectedPointsScore",
            "Sensor_OrbitStability"
        ]
    ):
        PluginLibrary.AddClassToGroup(sensor, sensor_name, {"simulator": sim, "platform_type": CWHAvailablePlatformTypes})



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This package implements the simulators used in the safe
autonomy environments.

Simulators are responsible for coordinating interactions between
the various objects involved in a simulation.

CoRL simulators act as a communication layer between CoRL and any environment
simulator you want to use.

`SafeRLSimulator` and its subclasses pair
[platforms](../platforms/index.md) with their backend simulation
entities to enable accurate environment interactions with respect to
implemented dynamics models. They also coordinate execution of
[done functions](../dones/index.md) to terminate training
episodes appropriately.
"""



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This package implements Runtime Assurance modules for use
within the CoRL framework.

Runtime Assurance (RTA) is a safety technique which intercepts
a control or action before it is executed and checks it against
a well-defined safety constraint to see if the action is safe.
If the action is safe, the RTA module allows it to execute.
If the action is deemed **unsafe** the RTA module provides an
alternative action which is guaranteed to be safe under the assumption
that the system is already in a safe state.
"""



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements the Reward Functions and Reward Validators specific to RTA.
"""
from collections import OrderedDict

from corl.libraries.state_dict import StateDict
from corl.rewards.reward_func_base import RewardFuncBase, RewardFuncBaseValidator


class RTAInterveningRewardValidator(RewardFuncBaseValidator):
    """
    A configuration validator for the RTAInterveningReward Reward Function.

    Attributes
    ----------
    scale : float
        Scalar value to adjust magnitude of the reward.
    """
    scale: float


class RTAInterveningReward(RewardFuncBase):
    """
    This reward function allocates reward based on if RTA is intervening or not.
    """

    def __init__(self, **kwargs):
        self.config: RTAInterveningRewardValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        return RTAInterveningRewardValidator

    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: StateDict,
        next_state: StateDict,
        observation_space: StateDict,
        observation_units: StateDict,
    ) -> float:

        reward = 0.
        if next_observation[self.config.agent_name]['RTAModule']['intervening']:
            reward = self.config.scale

        return reward



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

import matplotlib.pyplot as plt
import numpy as np
from corl.evaluation.runners.section_factories.plugins.platform_serializer import PlatformSerializer
from matplotlib.patches import FancyArrowPatch
from mpl_toolkits.mplot3d.axes3d import Axes3D
from mpl_toolkits.mplot3d.proj3d import proj_transform
from scipy.spatial.transform import Rotation

from safe_autonomy_sims.evaluation.animation.base_animation import BaseAnimationModule
from safe_autonomy_sims.evaluation.launch.serialize_cwh3d import SerializeCWH3D
from safe_autonomy_sims.simulators.inspection_simulator import InspectionPoints


class Arrow3D(FancyArrowPatch):
    """Make a 3d arrow
    Credit: https://gist.github.com/WetHat/1d6cd0f7309535311a539b42cccca89c
    """

    def __init__(self, x, y, z, dx, dy, dz, *args, **kwargs):
        super().__init__((0, 0), (0, 0), *args, **kwargs)
        self._xyz = (x, y, z)
        self._dxdydz = (dx, dy, dz)

    def draw(self, renderer):
        x1, y1, z1 = self._xyz
        dx, dy, dz = self._dxdydz
        x2, y2, z2 = (x1 + dx, y1 + dy, z1 + dz)

        xs, ys, _ = proj_transform((x1, x2), (y1, y2), (z1, z2), self.axes.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
        super().draw(renderer)

    def do_3d_projection(self, renderer=None):  # pylint:disable=unused-argument
        """3d projection"""
        x1, y1, z1 = self._xyz
        dx, dy, dz = self._dxdydz
        x2, y2, z2 = (x1 + dx, y1 + dy, z1 + dz)

        xs, ys, zs = proj_transform((x1, x2), (y1, y2), (z1, z2), self.axes.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))

        return np.min(zs)


def _arrow3D(ax, x, y, z, dx, dy, dz, *args, **kwargs):
    '''Add an 3d arrow to an `Axes3D` instance.'''

    arrow = Arrow3D(x, y, z, dx, dy, dz, *args, **kwargs)
    art = ax.add_artist(arrow)
    return art


setattr(Axes3D, 'arrow3D', _arrow3D)


class SixDofAnimation(BaseAnimationModule):
    """
    Module for making animations from a checkpoint for the 6 DoF inspection environment

    Parameters
    ----------
    expr_config_path: str
        The absolute path to the experiment config used in training
    task_config_path: str
        The absolute path to the task_config used in training
    platform_serializer_class: PlatformSerializer
        The PlatformSerializer subclass capable of saving the state of the Platforms used in the evaluation environment
    radius: float
        Radius of the sphere of inspection points
    num_points: int
        Number of inspection points
    """

    def __init__(
        self,
        expr_config_path: str = 'safe-autonomy-sims/configs/weighted-six-dof-inspection-v2/experiment.yml',
        task_config_path: str = 'safe-autonomy-sims/configs/weighted-six-dof-inspection-v2/task.yml',
        platform_serializer_class: PlatformSerializer = SerializeCWH3D,
        radius: float = 10.,
        num_points: int = 100,
        **kwargs
    ):
        super().__init__(
            expr_config_path=expr_config_path,
            task_config_path=task_config_path,
            platform_serializer_class=platform_serializer_class,
            **kwargs
        )
        self.init_sensor_vec = np.array([1., 0., 0.])
        self.temp_elements = []
        self.radius = radius
        self.num_points = num_points

    def get_data_from_dataframe(self, dataframe):
        positions = []
        velocities = []
        inspected_points = []
        sun_angles = []
        quaternions = []
        angular_velocities = []
        reaction_wheel_velocities = []
        temperatures = []
        energy = []

        for state in dataframe["ObservationVector"][0]:
            positions.append(state.data['ObserveSensor_Sensor_Position'].value['direct_observation'])
            velocities.append(state.data['ObserveSensor_Sensor_Velocity'].value['direct_observation'])
            inspected_points.append(state.data['ObserveSensor_Sensor_InspectedPoints'].value['direct_observation'])
            sun_angles.append(state.data['ObserveSensor_Sensor_SunAngle'].value['direct_observation'])
            quaternions.append(state.data['ObserveSensor_Sensor_Quaternion'].value['direct_observation'])
            angular_velocities.append(state.data['ObserveSensor_Sensor_AngularVelocity'].value['direct_observation'])
            try:
                reaction_wheel_velocities.append(state.data['ObserveSensor_Sensor_ReactionWheelVelocity'].value['direct_observation'])
                temperatures.append(state.data['ObserveSensor_Sensor_Temperature'].value['direct_observation'])
                energy.append(state.data['ObserveSensor_Sensor_Energy'].value['direct_observation'])
            except KeyError:
                pass

        data = {}

        data['positions'] = np.array(positions)
        data['velocities'] = np.array(velocities)
        data['inspected_points'] = np.array(inspected_points)
        data['sun_angles'] = np.array(sun_angles)
        data['quaternions'] = np.array(quaternions)
        data['angular_velocities'] = np.array(angular_velocities)
        data['reaction_wheel_velocities'] = np.array(reaction_wheel_velocities)
        data['temperatures'] = np.array(temperatures)
        data['energy'] = np.array(energy)

        control_vectors = []
        for control in dataframe["ControlVector"][0]:
            step_control = []
            for v in control.data.values():
                step_control.append(v.value[0])
            control_vectors.append(step_control)
        data['control_vectors'] = np.array(control_vectors)

        return data

    def setup_plots(self, data):
        return_plots = {}

        max_position = np.max(abs(data['positions']))
        max_xy_pos = np.max(abs(data['positions'][:, 0:2]))

        fig1 = plt.figure(figsize=(10, 5))
        plt.rcParams.update({'font.size': 16, 'text.usetex': True})
        ax1 = fig1.add_subplot(121, projection='3d')
        ax1.set_xlabel(r'$x$')
        ax1.set_ylabel(r'$y$')
        ax1.set_zlabel(r'$z$')
        ax1.set_xlim([-max_position * 1.1, max_position * 1.1])
        ax1.set_ylim([-max_position * 1.1, max_position * 1.1])
        ax1.set_zlim([-max_position * 1.1, max_position * 1.1])
        ax1.set_aspect("equal")

        ax2 = fig1.add_subplot(122)
        ax2.set_xlabel(r'$x$')
        ax2.set_ylabel(r'$y$')
        ax2.set_xlim([-max_xy_pos * 1.1, max_xy_pos * 1.1])
        ax2.set_ylim([-max_xy_pos * 1.1, max_xy_pos * 1.1])
        ax2.set_aspect("equal")
        ax2.grid(True)

        points = InspectionPoints.points_on_sphere_fibonacci(InspectionPoints, self.num_points, self.radius)
        for p in points:
            ax1.plot(p[0], p[1], p[2], 'g*', markersize=5)
            ax2.plot(p[0], p[1], 'g*', markersize=5)

        return_plots['ax1'] = ax1
        return_plots['ax2'] = ax2
        return return_plots

    def make_plots_in_loop(self, axes, data, i):
        max_position = np.max(abs(data['positions']))
        max_xy_pos = np.max(abs(data['positions'][:, 0:2]))

        if self.filetype != 'png':
            self.temp_elements = []

            rot = Rotation.from_quat(data['quaternions'][i])
            sensor_vec = rot.apply(self.init_sensor_vec)
            arrow = axes['ax1'].arrow3D(
                data['positions'][i][0],
                data['positions'][i][1],
                data['positions'][i][2],
                data['positions'][i][0] + sensor_vec[0] * max_position / 2,
                data['positions'][i][1] + sensor_vec[1] * max_position / 2,
                data['positions'][i][2] + sensor_vec[2] * max_position / 2,
                fc='c',
                ec='k',
                mutation_scale=20
            )
            arrow_xy = axes['ax2'].annotate(
                '',
                xy=(data['positions'][i][0] + sensor_vec[0] * max_xy_pos / 3, data['positions'][i][1] + sensor_vec[1] * max_xy_pos / 3),
                xytext=(data['positions'][i][0], data['positions'][i][1]),
                arrowprops={
                    'fc': 'c', 'ec': 'k'
                }
            )

            self.temp_elements.append(arrow)
            self.temp_elements.append(arrow_xy)

            plot_position = []
            plot_position.extend(
                axes['ax1'].plot(data['positions'][i][0], data['positions'][i][1], data['positions'][i][2], 'b*', markersize=15)
            )
            plot_position.extend(axes['ax2'].plot(data['positions'][i][0], data['positions'][i][1], 'b*', markersize=15))
            sun_vec = np.array([np.cos(data['sun_angles'][i]), -np.sin(data['sun_angles'][i]), 0.])
            point_inSunDir = max_position * sun_vec
            point_inSunDir_xy = max_xy_pos * sun_vec
            sun_line, = axes['ax1'].plot(
                [point_inSunDir[0], 0], [point_inSunDir[1], 0], [point_inSunDir[2], 0], linewidth=5, color='y', zorder=0
            )
            sun_line2, = axes['ax2'].plot([point_inSunDir_xy[0], 0], [point_inSunDir_xy[1], 0], linewidth=5, color='y', zorder=0)

            self.temp_elements.append(sun_line)
            self.temp_elements.append(sun_line2)
            self.temp_elements.append(plot_position)

        else:
            if i != 0:
                axes['ax1'].plot(
                    [data['positions'][i - 1][0], data['positions'][i][0]], [data['positions'][i - 1][1], data['positions'][i][1]],
                    [data['positions'][i - 1][2], data['positions'][i][2]],
                    'b',
                    linewidth=3
                )
                axes['ax2'].plot(
                    [data['positions'][i - 1][0], data['positions'][i][0]], [data['positions'][i - 1][1], data['positions'][i][1]],
                    'b',
                    linewidth=3
                )

    def remove_temp_elements(self, axes):
        for tmp in self.temp_elements:
            if isinstance(tmp, list):
                for p in tmp:
                    p.remove()
            else:
                tmp.remove()
        self.temp_elements = []

    def print_metrics(self, data):
        pass


if __name__ == '__main__':
    animation = SixDofAnimation(checkpoint_path='example_checkpoint')
    animation.make_animation(filetype='gif')



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module contains functions that define common terminal conditions across environments.
"""

import typing
from collections import OrderedDict

import gymnasium
import numpy as np
from corl.dones.done_func_base import DoneFuncBase, DoneFuncBaseValidator, DoneStatusCodes, SharedDoneFuncBase, SharedDoneFuncBaseValidator
from corl.libraries.environment_dict import DoneDict
from corl.libraries.state_dict import StateDict
from corl.simulators.common_platform_utils import get_platform_by_name


class TimeoutDoneValidator(DoneFuncBaseValidator):
    """
    A configuration validator for the TimeoutDoneFunction.

    Attributes
    ----------
    max_sim_time: float
        The maximum simulation time
    """
    max_sim_time: float


# TODO: remove redundant done func
class TimeoutDoneFunction(DoneFuncBase):
    """
    A done function that determines if the max episode time
    has been reached.

    Attributes
    ----------
    config: TimeoutDoneValidator
        The function's validated configuration parameters
    """

    def __init__(self, **kwargs) -> None:
        self.config: TimeoutDoneValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        """
        Returns the validator for this done function.

        Returns
        -------
        TimeoutDoneValidator
            Config validator for the TimeoutDoneFunction.
        """
        return TimeoutDoneValidator

    def __call__(
        self,
        observation,
        action,
        next_observation,
        next_state,
        observation_space: gymnasium.spaces.dict.Dict,
        observation_units: gymnasium.spaces.dict.Dict,
    ) -> bool:
        """
        Parameters
        ----------
        observation : np.ndarray
            np.ndarray describing the current observation
        action : np.ndarray
            np.ndarray describing the current action
        next_observation : np.ndarray
            np.ndarray describing the incoming observation
        next_state : np.ndarray
            np.ndarray describing the incoming state

        Returns
        -------
        done : bool
            the done condition for the current agent.
        """

        # get sim time
        platform = get_platform_by_name(next_state, self.config.platform_name)
        sim_time = platform.sim_time

        done = sim_time >= self.config.max_sim_time

        if done:
            next_state.episode_state[self.config.platform_name][self.name] = DoneStatusCodes.LOSE

        return done


class CollisionDoneFunctionValidator(SharedDoneFuncBaseValidator):
    """
    A configuration validator for the CollisionDoneFunction.

    Attributes
    ----------
    safety_constraint : float
        The minimum radial distance between spacecrafts that must be maintained in order to avoid a collision.
    """
    safety_constraint: float = 0.5  # meters


class CollisionDoneFunction(SharedDoneFuncBase):
    """
    A done function that determines if an agent's spacecraft
    has collided with another agent's spacecraft in the environemt.

    Attributes
    ----------
    config: CollisionDoneFunctionValidator
        The function's validated configuration parameters
    """

    @staticmethod
    def get_validator() -> typing.Type[SharedDoneFuncBaseValidator]:
        """
        Returns the validator for this done function.

        Returns
        -------
        CollisionDoneFunctionValidator
            Config validator for the CollisionDoneFunction.
        """
        return CollisionDoneFunctionValidator

    def __call__(
        self,
        observation: OrderedDict,
        action: OrderedDict,
        next_observation: OrderedDict,
        next_state: StateDict,
        observation_space: gymnasium.spaces.dict.Dict,
        observation_units: gymnasium.spaces.dict.Dict,
        local_dones: DoneDict,
        local_done_info: OrderedDict
    ) -> DoneDict:
        """
        Parameters
        ----------
        observation : np.ndarray
            np.ndarray describing the current observation
        action : np.ndarray
            np.ndarray describing the current action
        next_observation : np.ndarray
            np.ndarray describing the incoming observation
        next_state : np.ndarray
            np.ndarray describing the incoming state
        observation_space : gymnasium.spaces.dict.Dict
            The agent observation space.
        observation_units : gymnasium.spaces.dict.Dict
            The units of the observations in the observation space. This may be None.
        local_dones: DoneDict
            DoneDict containing name to boolean KVPs representing done statuses of each agent
        local_done_info: OrderedDict
            An OrderedDict containing nested OrderedDicts of done function to done status KVPs for each agent

        Returns
        -------
        done : DoneDict
            Dictionary containing the done condition for each agent.
        """

        # get list of spacecrafts
        platform_names = list(local_dones.keys())
        try:
            platform_names.remove("__all__")
        except ValueError:
            # TODO: remove try catch, if not necessary
            # print(err)
            pass

        # populate DoneDict
        done = DoneDict()
        for name in local_dones.keys():
            done[name] = False

        # check if any platform violates boundaries of other platforms
        while len(platform_names) > 1:
            name = platform_names.pop()
            platform = get_platform_by_name(next_state, name)
            position = platform.position

            for other_platform_name in platform_names:

                if local_dones[other_platform_name]:
                    # skip if other_agent is done
                    continue

                # check location against location of other platforms for boundary violation
                other_platform = get_platform_by_name(next_state, other_platform_name)
                other_platform_position = other_platform.position

                radial_distance = np.linalg.norm(np.array(position) - np.array(other_platform_position))

                if radial_distance < self.config.safety_constraint:
                    # collision detected. stop loop and end episode
                    for k in local_dones.keys():
                        done[k] = True
                    # break
                    return done
        return done


class MultiagentSuccessDoneFunctionValidator(SharedDoneFuncBaseValidator):
    """
    A configuration validator for the MultiagentSuccessfulDockingDoneFunction.

    Attributes
    ----------
    success_function_name : str
        The name of the successful docking function, which this
        function will reference to ensure all agents have reached a
        DoneStatusCodes.WIN state before ending the episode.
    """
    success_function_name: str = "RejoinSuccessDone"


class MultiagentSuccessDoneFunction(SharedDoneFuncBase):
    """
    This done function determines whether every agent in the
    environment has reached a specified successful done condition.

    Attributes
    ----------
    config: MultiagentSuccessDoneFunctionValidator
        The function's validated configuration parameters
    """

    @staticmethod
    def get_validator() -> typing.Type[SharedDoneFuncBaseValidator]:
        """
        Returns the validator for this done function.

        Returns
        -------
        MultiagentSuccessDoneFunctionValidator
            Config validator for the MultiagentSuccessDoneFunction.
        """
        return MultiagentSuccessDoneFunctionValidator

    def __call__(
        self,
        observation: OrderedDict,
        action: OrderedDict,
        next_observation: OrderedDict,
        next_state: StateDict,
        observation_space: gymnasium.spaces.dict.Dict,
        observation_units: gymnasium.spaces.dict.Dict,
        local_dones: DoneDict,
        local_done_info: OrderedDict
    ) -> DoneDict:
        """
        Parameters
        ----------
        observation : np.ndarray
            np.ndarray describing the current observation
        action : np.ndarray
            np.ndarray describing the current action
        next_observation : np.ndarray
            np.ndarray describing the incoming observation
        next_state : np.ndarray
            np.ndarray describing the incoming state
        observation_space : gymnasium.spaces.dict.Dict
            The agent observation space.
        observation_units : gymnasium.spaces.dict.Dict
            The units of the observations in the observation space. This may be None.
        local_dones: DoneDict
            DoneDict containing name to boolean KVPs representing done statuses of each agent
        local_done_info: OrderedDict
            An OrderedDict containing nested OrderedDicts of done function to done status KVPs for each agent

        Returns
        -------
        done : DoneDict
            Dictionary containing the done condition for each agent.
        """

        # populate DoneDict
        done = DoneDict()
        for name in local_dones.keys():
            done[name] = False

        for platform_name in local_done_info.keys():
            if self.config.success_function_name in next_state.episode_state[platform_name]:
                # docking kvp exists
                if next_state.episode_state[platform_name][self.config.success_function_name] != DoneStatusCodes.WIN:
                    # agent failed to reach WIN condition
                    return done
            else:
                # agent has not reached done condition
                return done

        # all agents have succeeded, set all dones to True
        for k in local_dones.keys():
            done[k] = True
        return done


class SetAllDoneFunction(SharedDoneFuncBase):
    """
    A done function that sets all agents to done if one agent
    is found to be done.
    """

    def __call__(
        self,
        observation: OrderedDict,
        action: OrderedDict,
        next_observation: OrderedDict,
        next_state: StateDict,
        observation_space: gymnasium.spaces.dict.Dict,
        observation_units: gymnasium.spaces.dict.Dict,
        local_dones: DoneDict,
        local_done_info: OrderedDict
    ) -> DoneDict:
        """
        Parameters
        ----------
        observation : np.ndarray
            np.ndarray describing the current observation
        action : np.ndarray
            np.ndarray describing the current action
        next_observation : np.ndarray
            np.ndarray describing the incoming observation
        next_state : np.ndarray
            np.ndarray describing the incoming state
        observation_space : gymnasium.spaces.dict.Dict
            The agent observation space.
        observation_units : gymnasium.spaces.dict.Dict
            The units of the observations in the observation space. This may be None.
        local_dones: DoneDict
            DoneDict containing name to boolean KVPs representing done statuses of each agent
        local_done_info: OrderedDict
            An OrderedDict containing nested OrderedDicts of done function to done status KVPs for each agent

        Returns
        -------
        done : DoneDict
            Dictionary containing the done condition for each agent.
        """

        # get list of spacecrafts
        platform_names = list(local_dones.keys())
        # populate DoneDict
        dones = DoneDict()

        done = any(local_dones.values())

        for name in platform_names:
            dones[name] = done

        return dones



from gymnasium.envs.registration import register


# ---- single agent environments ----

register(
    id="Docking-v0",
    entry_point="safe_autonomy_sims.gym.docking.docking_v0:DockingEnv",
)

register(
    id="Inspection-v0",
    entry_point="safe_autonomy_sims.gym.inspection.inspection_v0:InspectionEnv",
)

register(
    id="WeightedInspection-v0",
    entry_point="safe_autonomy_sims.gym.inspection.weighted_inspection_v0:WeightedInspectionEnv",
)

register(
    id="SixDofInspection-v0",
    entry_point="safe_autonomy_sims.gym.inspection.sixdof_inspection_v0:WeightedSixDofInspectionEnv",
)



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module defines a python API for running CoRL's Evaluation Framework.

It also defines helper functions to support the reuse
of training configs, evaluation of multiple trained policies,
and creation of comparative plots of chosen Metrics. These functions
streamline visualization and analysis of comparative RL test assays.
"""

# pylint: disable=E0401

import re
import sys
import typing

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from corl.evaluation.api import evaluate
from corl.evaluation.api_utils import add_required_metrics, construct_dataframe, extract_metadata_from_result_file, get_checkpoints_paths
from corl.evaluation.evaluation_artifacts import (
    EvaluationArtifact_EvaluationOutcome,
    EvaluationArtifact_Metrics,
    EvaluationArtifact_Visualization,
)
from corl.evaluation.launchers import launch_generate_metrics, launch_visualize
from corl.evaluation.recording.folder import FolderRecord
from corl.evaluation.runners.section_factories.plugins.platform_serializer import PlatformSerializer
from corl.evaluation.runners.section_factories.test_cases.default_strategy import DefaultStrategy
from corl.evaluation.visualization.print import Print


def run_ablation_study(
    experiment_output_paths_map: dict[str, str],
    task_config_path: str,
    experiment_config_path: str,
    launch_dir_of_experiment: str,
    metrics_config: dict[str, dict | list],
    platfrom_serializer_class: PlatformSerializer,
    evaluation_ouput_dir: str,
    plot_output_path: str | None = None,
    plot_config: dict | None = None,
    create_plot: bool = False,
    test_case_manager_config: dict | None = None,
    rl_algorithm_name: str | None = None,
    num_workers: int = 1
) -> pd.DataFrame:  # pylint:disable=too-many-locals
    """
    This function serves as a high level interface for users conducting ablation studies using CoRL. It is
    responsible for processing a collection of CoRL training outputs (experiment output directories).
    This includes iteratively loading policies from checkpoints, running those policies in evaluation episodes,
    collecting data, analyzing data with Metrics, and optionally visualizing calculated Metrics in a plot.

    Parameters
    ----------
    experiment_output_paths_map: dict[str, str]
        A map of experiment name string keys to experiment output directory path strings
    task_config_path: str
        The absolute path to the task config yaml file used in training
    experiment_config_path: str
        The absolute path to the experiment config used in training
    launch_dir_of_experiment: str
        Paths in the experiment config are relative to the directory the corl.train_rl module is intended to be launched from.
        This string captures the path to the intended launch location of the experiment under evaluation.
    metrics_config: dict
        The nested structure defining the Metrics to be instantiated and calculated from Evaluation Episode data
    platfrom_serializer_class: PlatformSerializer
        The class object capable of storing data specific to the Platform type used in training
    plot_output_path: str
        The absolute path to the location where the resulting plot will be saved
    plot_config: dict
        The kwargs that will be passed to seaborn's lineplot method
    create_plot: bool
        A boolean to determine whether or not to create the seaborn plot. If True, the plot is created and saved
    test_case_manager_config: dict
        The kwargs passed to the TestCaseManager constructor. This must define the TestCaseStrategy class and its config

    Returns
    -------
    data: pandas.DataFrame
        The DataFrame containing Metric data from all provided experiments collected during evaluation
    """

    # handle default TestCaseManager config
    if test_case_manager_config is None:
        test_case_manager_config = {"type": f"{DefaultStrategy.__module__}.{DefaultStrategy.__name__}", "config": {"num_test_cases": 3}}
    # handle default plot config
    if plot_config is None:
        plot_config = {}

    # add required metrics
    metrics_config = add_required_metrics(metrics_config)

    # evaluate provided trained policies' checkpoints

    # construct keys by appending experiment name and trial index together
    trial_to_eval_results_map = {}

    for experiment_name, trial_dir_path in experiment_output_paths_map.items():

        # if single training output provided, wrap in list
        trial_dirs = [trial_dir_path] if not isinstance(trial_dir_path, list) else trial_dir_path
        for trial_index, trial_dir_path in enumerate(trial_dirs):

            # create trial key
            trial_key = f"{experiment_name}__{trial_index}"

            checkpoint_paths, output_paths = get_checkpoints_paths(trial_dir_path, trial_key, evaluation_ouput_dir)
            metadata = extract_metadata_from_result_file(trial_dir_path)

            # temporarily index experiments listed under same name
            trial_to_eval_results_map[trial_key] = {"output_paths": output_paths, "metadata": metadata}
            run_evaluations(
                task_config_path,
                experiment_config_path,
                launch_dir_of_experiment,
                metrics_config,
                checkpoint_paths,
                output_paths,
                platfrom_serializer_class,
                test_case_manager_config=test_case_manager_config,
                visualize_metrics=False,
                rl_algorithm_name=rl_algorithm_name,
                experiment_name=experiment_name,
                num_workers=num_workers,
            )

    # create sample complexity plot
    data = construct_dataframe(trial_to_eval_results_map, metrics_config)

    # TODO: for metric in metric_config: (plot every given metric?)
    if create_plot:
        assert plot_output_path is not None, "'plot_output_path' must be defined in oder to save created plot"
        create_sample_complexity_plot(data, plot_output_path, **plot_config)

    return data


def generate_metrics(evaluate_output_path: str, metrics_config: dict[str, dict | list]):
    """
    This function is responsible for instantiating necessary arguments and then launching the second stage of CoRL's Evaluation Framework.

    Parameters
    ----------
    evaluate_output_path: str
        The absolute path to the directory in which evaluation episodes' data was saved (from the initial 'evaluate' step of the
        Evaluation Framework)
    metrics_config: dict
        The nested structure defining the Metrics to be instantiated and calculated from Evaluation Episode data
    """

    # define constructor args
    location = FolderRecord(absolute_path=evaluate_output_path)

    # TODO: enable evaluation without alerts
    alerts_config = {
        "world": [
            {
                "name": "Short Episodes",
                "metric": "rate_of_runs_lt_5steps",
                "scope": "evaluation",
                "thresholds": [{
                    "type": "warning", "condition": {
                        "operator": ">", "lhs": 0
                    }
                }]
            }
        ]
    }

    raise_error_on_alert = True

    # instantiate eval objects
    outcome = EvaluationArtifact_EvaluationOutcome(location=location)
    metrics = EvaluationArtifact_Metrics(location=evaluate_output_path)

    # construct namespace dict
    namespace = {
        "artifact_evaluation_outcome": outcome,
        "artifact_metrics": metrics,
        "metrics_config": metrics_config,
        "alerts_config": alerts_config,
        "raise_on_error_alert": raise_error_on_alert
    }

    launch_generate_metrics.main(namespace)


def visualize(evaluate_output_path: str):
    """
    This function is responsible for instantiating necessary arguments and then launching the third stage of CoRL's Evaluation Framework.

    Parameters
    ----------
    evaluate_output_path: str
        The absolute path to the directory in which evaluation episodes' data was saved (from the initial 'evaluate' step
        of the Evaluation Framework)
    """

    artifact_metrics = EvaluationArtifact_Metrics(location=evaluate_output_path)
    artifact_visualization = EvaluationArtifact_Visualization(location=evaluate_output_path)
    visualizations = [Print(event_table_print=True)]

    namespace = {"artifact_metrics": artifact_metrics, "artifact_visualization": artifact_visualization, "visualizations": visualizations}

    launch_visualize.main(namespace)


def run_evaluations(
    task_config_path: str,
    experiment_config_path: str,
    launch_dir_of_experiment: str,
    metrics_config: dict[str, dict | list],
    checkpoint_paths: list,
    output_paths: list,
    platfrom_serializer_class: PlatformSerializer,
    test_case_manager_config: dict | None = None,
    visualize_metrics: bool = False,
    rl_algorithm_name: str | None = None,
    experiment_name: str = "",
    num_workers: int = 1
):
    """
    This function is responsible for taking a list of checkpoint paths and iteratively running them through
    each stage of the Evaluation Framework (running Evaluation Episodes, processing results to generate Metrics,
    and optionally visualizing those Metrics).

    Parameters
    ----------
    task_config_path: str
        The absolute path to the directory in which evaluation episode(s) data will be saved
    experiment_config_path: str
        The absolute path to the experiment config used in training
    launch_dir_of_experiment: str
        Paths in the experiment config are relative to the directory the corl.train_rl module is intended to be launched from.
        This string captures the path to the intended launch location of the experiment under evaluation.
    metrics_config: dict
        The nested structure defining the Metrics to be instantiated and calculated from Evaluation Episode data
    checkpoint_paths: list
        A list of path strings to each checkpoint, typically checkpoints are from a single training job
    output_paths: list
        A list of unique path strings at which each checkpoint's Evaluation Episodes data and Metrics will be stored.
        Must match the length of the checkpoint_paths list
    platfrom_serializer_class: PlatformSerializer
        The class object capable of storing data specific to the Platform type used in training
    test_case_manager_config: dict
        The kwargs passed to the TestCaseManager constructor. This must define the TestCaseStrategy class and its config
    visualize_metrics: bool
        A boolean to determine whether or not to run the Evaluation Framework's Visualize stage.
        If True, visualize is called.
    rl_algorithm_name: str
        The name of the RL algorithm to use in Evaluation Episodes
    experiment_name: str
        The name of the experiment. Used for standard output progress updates.
    num_workers: int
        The number of ray workers to use when running evaluation episodes in parallel.
    """

    kwargs: dict[str, typing.Any] = {}
    if test_case_manager_config:
        kwargs['test_case_manager_config'] = test_case_manager_config

    # rl algorithm
    if rl_algorithm_name:
        kwargs['rl_algorithm_name'] = rl_algorithm_name

    kwargs["num_workers"] = num_workers

    # run sequence of evaluation
    for index, ckpt_path in enumerate(checkpoint_paths):
        # print progress
        ckpt_num_regex = re.search(r'(checkpoint_)\w+', ckpt_path)
        ckpt_num = ckpt_path[ckpt_num_regex.start():ckpt_num_regex.end()]  # type: ignore
        if experiment_name:
            print("\nExperiment: " + experiment_name)
        print("Evaluating " + ckpt_num + ". " + str(index + 1) + " of " + str(len(checkpoint_paths)) + " checkpoints.")

        # run evaluation episodes
        try:
            evaluate(
                task_config_path,
                ckpt_path,
                output_paths[index],
                experiment_config_path,
                launch_dir_of_experiment,
                platfrom_serializer_class,
                **kwargs
            )
        except SystemExit:
            print(sys.exc_info()[0])

        # generate evaluation metrics
        generate_metrics(output_paths[index], metrics_config)

        if visualize_metrics:
            # generate visualizations
            visualize(output_paths[index])


def run_one_evaluation(
    task_config_path: str,
    experiment_config_path: str,
    launch_dir_of_experiment: str,
    metrics_config: dict[str, dict | list],
    checkpoint_path: str,
    platfrom_serializer_class: PlatformSerializer,
    test_case_manager_config: dict | None = None,
    rl_algorithm_name: str | None = None,
    num_workers: int = 1
) -> pd.DataFrame:
    """
    This function is responsible for taking a single checkpoint path and running it through
    each stage of the Evaluation Framework (running Evaluation Episodes, processing results
    to generate Metrics, and optionally visualizing those Metrics).

    Parameters
    ----------
    task_config_path: str
        The absolute path to the directory in which evaluation episode(s) data will be saved
    experiment_config_path: str
        The absolute path to the experiment config used in training
    launch_dir_of_experiment: str
        Paths in the experiment config are relative to the directory the corl.train_rl module is intended to be launched from.
        This string captures the path to the intended launch location of the experiment under evaluation.
    metrics_config: dict
        The nested structure defining the Metrics to be instantiated and calculated from Evaluation Episode data
    checkpoint_path: str
        Path strings to the checkpoint
    platfrom_serializer_class: PlatformSerializer
        The class object capable of storing data specific to the Platform type used in training
    test_case_manager_config: dict
        The kwargs passed to the TestCaseManager constructor. This must define the TestCaseStrategy class and its config
    rl_algorithm_name: str
        The name of the RL algorithm to use in Evaluation Episodes
    num_workers: int
        The number of ray workers to use when running evaluation episodes in parallel.
    """

    kwargs: dict[str, typing.Any] = {}
    if test_case_manager_config:
        kwargs['test_case_manager_config'] = test_case_manager_config

    # rl algorithm
    if rl_algorithm_name:
        kwargs['rl_algorithm_name'] = rl_algorithm_name

    kwargs["num_workers"] = num_workers

    metrics_config = add_required_metrics(metrics_config)

    checkpoint_path += "/policies/{}/policy_state.pkl"
    exp_name = 'single_episode__0__0'
    output_path = '/tmp/eval_results/' + exp_name

    # run evaluation episodes
    try:
        evaluate(
            task_config_path,
            checkpoint_path,
            output_path,
            experiment_config_path,
            launch_dir_of_experiment,
            platfrom_serializer_class,
            **kwargs
        )
    except SystemExit:
        print(sys.exc_info()[0])

    # generate evaluation metrics
    generate_metrics(output_path, metrics_config)

    experiment_to_eval_results_map = {}
    experiment_to_eval_results_map[exp_name] = {"output_paths": [output_path], "metadata": pd.DataFrame({'training_iteration': [0.]})}
    data = construct_dataframe(experiment_to_eval_results_map, metrics_config)
    return data


def create_sample_complexity_plot(
    data: pd.DataFrame,
    plot_output_file: str,
    xaxis="timesteps_total",
    yaxis="TotalReward",
    hue="agent",
    errorbar=("ci", 95),
    xmax=None,
    ylim=None,
    **kwargs,
) -> matplotlib.axis.Axis:
    """
    This function is responsible for creating a comparative sample complexity plot with seaborn. The function takes in
    a pandas.DataFrame of Evaluation results and specifications of kwargs to determine x and y axis and confidence
    intervals and saves the resulting plot to the desired location.

    Parameters
    ----------
    data: pd.DataFrame
        A DataFrame containing Metric values collected for each checkpoint of each experiment by row
    plot_output_file: str
        The location at which the plot will be saved
    xaxis: str
        The x axis of the plot. Must be a column in the 'data' DataFrame. Currently 'iterations',
        'num_episodes', and 'num_interactions' are supported
    yaxis: str
        The y axis of the plot. Must be a column in the 'data' DataFrame, typically the name of a Metric that was
        included in the metrics_config during Evaluation.
    hue: str
        The column of the 'data' DataFrame on which to group data
    errorbar: tuple
        The confidence interval to be displayed on the plot
    xmax: int
        The maximum value displayed on the x axis
    ylim: int
        The maximum value displayed on the y axis
    kwargs
        Additional kwargs to pass to seaborn's lineplot method

    Returns
    -------
    plot: matplotlib.axis.Axis
        An axis containing the resulting plot
    """
    # create seaborn plot
    # TODO: upgrade seaborn to 0.12.0 and swap deprecated 'ci' for 'errorbar' kwarg: errorbar=('sd', 1) or (''ci', 95)
    # TODO: add smooth kwarg

    plt.clf()

    sns.set(style="darkgrid", font_scale=1.5)

    # feed Dataframe to seaborn to create labelled plot
    plot = sns.lineplot(data=data, x=xaxis, y=yaxis, hue=hue, errorbar=errorbar, **kwargs)

    # format plot

    plt.legend(loc="best").set_draggable(True)

    if xaxis == "timesteps_total":
        plt.xlabel("Timesteps")
    if xaxis == "episodes_total":
        plt.xlabel("Episodes")
    if xaxis == "training_iteration":
        plt.xlabel("Iterations")
    if yaxis == "TotalReward":
        plt.ylabel("Average Reward")

    if xmax is None:
        xmax = np.max(np.asarray(data[xaxis]))
    plt.xlim(right=xmax)

    if ylim is not None:
        plt.ylim(ylim)

    xscale = xmax > 5e3
    if xscale:
        # Just some formatting niceness: x-axis scale in scientific notation if max x is large
        plt.ticklabel_format(style="sci", axis="x", scilimits=(0, 0))

    plt.tight_layout(pad=0.5)

    # save plot

    plot.get_figure().savefig(plot_output_file)

    return plot



"""Utility functions for the inspection environment"""

import numpy as np
import safe_autonomy_simulation.entities as e
import safe_autonomy_simulation.sims.inspection as sim


def polar_to_cartesian(r: float, theta: float, phi: float) -> np.ndarray:
    """Convert polar coordinates to cartesian coordinates.

    Parameters
    ----------
    r : float
        radial distance
    theta : float
        azimuthal angle
    phi : float
        polar angle

    Returns
    -------
    np.ndarray
        cartesian coordinates
    """
    x = r * np.sin(phi) * np.cos(theta)
    y = r * np.sin(phi) * np.sin(theta)
    z = r * np.cos(phi)
    return np.array([x, y, z])


def rel_dist(pos1: np.ndarray, pos2: np.ndarray) -> float:
    """The relative distance between two positions.

    Parameters
    ----------
    pos1 : np.ndarray
        the first position
    pos2 : np.ndarray
        the second position

    Returns
    -------
    float
        the euclidean distance between the two positions
    """
    rel_d = np.linalg.norm(pos1 - pos2)
    return rel_d


def rel_vel(vel1: np.ndarray, vel2: np.ndarray) -> float:
    """The relative velocity between two velocities.

    Parameters
    ----------
    vel1 : np.ndarray
        the first velocity
    vel2 : np.ndarray
        the second velocity

    Returns
    -------
    float
        the relative velocity between the two velocities
    """
    rel_v = np.linalg.norm(vel1 - vel2)
    return rel_v


def delta_v(v: np.ndarray, prev_v: np.ndarray) -> np.ndarray:
    """The change in velocity

    Parameters
    ----------
    v : np.ndarray
        the current velocity
    prev_v : np.ndarray
        the previous velocity

    Returns
    -------
    float
        the change in velocity
    """
    v_norm = np.linalg.norm(v)
    prev_v_norm = np.linalg.norm(prev_v)
    return v_norm - prev_v_norm


def v_limit(
    chief_pos: np.ndarray,
    deputy_pos: np.ndarray,
    a: float = 2.0,
    n: float = 0.001027,
    v_max: float = 0.2,
    docking_radius: float = 0.5,
):
    """A linear velocity limit based on the distance between the chief
    and the deputy.

    $v_{limit} = v_{max} + an(d-r)$

    where:
    * $v_max$ is a maximum allowable velocity
    * $a$ is the slope of the linear velocity limit
    * $d$ is the distance between the chief and the deputy
    * $r$ is the radius of the docking region around the chief

    Parameters
    ----------
    chief_pos : np.ndarray
        the position of the chief
    deputy_pos : np.ndarray
        the position of the deputy
    a : float, optional
        slope of the linear velocity limit, by default 2.0
    n : float, optional
        deputy mean motion, by default 0.001027
    v_max : float, optional
        maximum allowable velocity, by default 0.2
    docking_radius : float, optional
        the radius of the docking region around the chief, by default 0.5

    Returns
    -------
    float
        velocity limit
    """
    v_limit = v_max + (
        a * n * (rel_dist(pos1=chief_pos, pos2=deputy_pos) - docking_radius)
    )
    return v_limit


def closest_fft_distance(
    chief: sim.Target, deputy: sim.Inspector, n: float = 0.001027, time_step: int = 1
) -> float:
    """
    Get the closest Free Flight Trajectory (FFT) distance between the deputy
    and the chief over one orbit using closed form CWH dynamics to calculate
    future states

    Parameters
    ----------
    chief : sim.Target
        chief spacecraft under inspection
    deputy : sim.Inspector
        deputy spacecraft performing inspection
    n: float, optional
        orbital mean motion of Hill's reference frame's circular orbit in rad/s, by default 0.001027
    time_step: int
        time step in seconds to calculate FFT over, by default 1

    Returns
    -------
    float
        closest relative distance between deputy and chief during the FFT
    """

    def get_pos(platform: e.PhysicalEntity, t: int):
        x = (
            (4 - 3 * np.cos(n * t)) * platform.x
            + np.sin(n * t) * platform.x_dot / n
            + 2 / n * (1 - np.cos(n * t)) * platform.y_dot
        )
        y = (
            6 * (np.sin(n * t) - n * t) * platform.x
            + platform.y
            - 2 / n * (1 - np.cos(n * t)) * platform.x_dot
            + (4 * np.sin(n * t) - 3 * n * t) * platform.y_dot / n
        )
        z = platform.z * np.cos(n * t) + platform.z_dot / n * np.sin(n * t)
        return np.array([x, y, z])

    distances = []
    times = np.arange(0, 2 * np.pi / n, time_step)
    for time in times:
        dep_pos = get_pos(platform=deputy, t=time)
        chief_pos = get_pos(platform=chief, t=time)
        distances.append(np.linalg.norm(chief_pos - dep_pos))
    return float(min(distances))



"""A gymnasium environment for training agents to dock a deputy spacecraft onto a chief spacecraft"""

import typing
import numpy as np
import gymnasium as gym
import safe_autonomy_simulation.sims.inspection as sim
import safe_autonomy_sims.gym.inspection.reward as r
import safe_autonomy_sims.gym.inspection.utils as utils


class InspectionEnv(gym.Env):
    r"""
    ## Description

    This environment represents a spacecraft inspection problem
    The goal is for a single deputy spacecraft to navigate around and inspect
    the entire surface of a chief spacecraft.

    The chief is covered in 100 inspection points that the agent must
    observe while they are illuminated by the moving sun. The optimal
    policy will inspect all 100 points within 2 revolutions of the sun
    while using as little fuel as possible.
    
    In this inspection environment, the agent only controls its translational
    motion and is always assumed to be pointing at the chief spacecraft.
    
    __Note: the policy selects a new action every 10 seconds__

    ## Action Space

    Actions are thruster force values for each of the 3 bi-directional thrusters
    on the x-, y-, and z-axis of the deputy spacecraft's body frame.

    The action space is a `Box(-1, 1, shape=(3,))` where each value in the array
    represents the force applied to the x-, y-, and z-axis of the deputy spacecraft.

    | Index | Action        | Control Min | Control Max | Type (units) |
    |-------|---------------|-------------|-------------|--------------|
    | 0     | x-axis thrust | -1          | 1           | Force (N)    |
    | 1     | y-axis thrust | -1          | 1           | Force (N)    |
    | 2     | z-axis thrust | -1          | 1           | Force (N)    |

    ## Observation Space

    At each timestep, the agent receives the observation, $o = [x, y, z, v_x, v_y, v_z, \theta_{sun}, n, x_{ups}, y_{ups}, z_{ups}]$, where:

    * $x, y,$ and $z$ represent the deputy's position in the Hill's frame,
    * $v_x, v_y,$ and $v_z$ represent the deputy's directional velocity in the Hill's frame,
    * $\theta_{sun}$ is the angle of the sun,
    * $n$ is the number of points that have been inspected so far and,
    * $x_{ups}, y_{ups},$ and $z_{ups}$ are the unit vectors pointing to the nearest large cluster of unispected points (found via k-means)

    The observation space is a `Box` with the following bounds:

    | Index | Observation                                            | Min  | Max    | Type (units)   |
    |-------|--------------------------------------------------------|------|--------|----------------|
    | 0     | x position of the deputy in Hill's frame               | -inf | inf    | position (m)   |
    | 1     | y position of the deputy in Hill's frame               | -inf | inf    | position (m)   |
    | 2     | z position of the deputy in Hill's frame               | -inf | inf    | position (m)   |
    | 3     | x velocity of the deputy in Hill's frame               | -inf | inf    | velocity (m/s) |
    | 4     | y velocity of the deputy in Hill's frame               | -inf | inf    | velocity (m/s) |
    | 5     | z velocity of the deputy in Hill's frame               | -inf | inf    | velocity (m/s) |
    | 6     | angle of the sun                                       | 0    | $2\pi$ | angle (rad)    |
    | 7     | number of points inspected                             | 0    | 100    | count          |
    | 8     | x component of unit vector pointing to nearest cluster | -1   | 1      | scalar         |
    | 9     | y component of unit vector pointing to nearest cluster | -1   | 1      | scalar         |
    | 10    | z component of unit vector pointing to nearest cluster | -1   | 1      | scalar         |

    ## State Transition Dynamics

    The relative motion between the deputy and chief is represented by
    linearized Clohessy-Wiltshire equations [[1]](#1), given by

    $$
        \dot{\boldsymbol{x}} = A {\boldsymbol{x}} + B\boldsymbol{u},
    $$

    where:
    * state $\boldsymbol{x}=[x,y,z,\dot{x},\dot{y},\dot{z}]^T \in \mathcal{X}=\mathbb{R}^6$
    * control $\boldsymbol{u}= [F_x,F_y,F_z]^T \in \mathcal{U} = [-1N, 1N]^3$

    * $$
        A =
    \begin{bmatrix}
    0 & 0 & 0 & 1 & 0 & 0 \\
    0 & 0 & 0 & 0 & 1 & 0 \\
    0 & 0 & 0 & 0 & 0 & 1 \\
    3n^2 & 0 & 0 & 0 & 2n & 0 \\
    0 & 0 & 0 & -2n & 0 & 0 \\
    0 & 0 & -n^2 & 0 & 0 & 0 \\
    \end{bmatrix},
        B =
    \begin{bmatrix}
    0 & 0 & 0 \\
    0 & 0 & 0 \\
    0 & 0 & 0 \\
    \frac{1}{m} & 0 & 0 \\
    0 & \frac{1}{m} & 0 \\
    0 & 0 & \frac{1}{m} \\
    \end{bmatrix},
    $$

    * mean motion constant $n = 0.001027 rad/s$

    ## Rewards

    The reward $r_t$ at each time step is the sum of the following terms:

    * $r_t += 0.1(num\_inspected\_points_t - num\_inspected\_points_{t-1})$
        * a dense reward for observing new points
    * $r $= 1$ if $num\_inspected\_points_i == 100$, else 0
        * a sparse reward for successfully inspecting all points
    * $r += -1$ if $radius < crash\_region\_radius$, else 0
        * a sparse reward punishing the agent for crashing with the chief
    * $r += -0.1||\boldsymbol{u}||$
        * a dense reward for minimizing fuel cost
        * fuel cost is represented as the norm of the control input (action)

    ## Starting State

    At the start of an episode, the state is randomly initialized with the following conditions:

    * chief $(x,y,z)$ = $(0, 0, 0)$
    * chief radius = $10 m$
    * chief # of points = $100$
    * deputy position $(x, y, z)$ is converted after randomly selecting the position in polar notation $(r, \phi, \psi)$ using a uniform distribution with
        * $r \in [50, 100] m$
        * $\psi \in [0, 2\pi] rad$
        * $\phi \in [-\pi/2, \pi/2] rad$
        * $x = r \cos{\psi} \cos{\phi}$
        * $y = r \sin{\psi} \cos{\phi}$
        * $z = r \sin{\phi}$
    * deputy $(v_x, v_y, v_z)$ is converted after randomly selecting the velocity in polar notation $(r, \phi, \psi)$ using a Gaussian distribution with
        * $v \in [0, 0.8] * v_{\rm max}, \quad v_{\rm max} = 2nr$ m/s
        * $\psi \in [0, 2\pi] rad$
        * $\phi \in [-\pi/2, \pi/2] rad$
        * $v_x = v \cos{\psi} \cos{\phi}$
        * $v_y = v \sin{\psi} \cos{\phi}$
        * $v_z = v \sin{\phi}$
    * deputy camera parameters:
        * field of view = $\pi$ rad
        * focal length = $1 m$
    * Initial sun angle is randomly selected using a uniform distribution
        * $\theta_{sun} \in [0, 2\pi] rad$
    * Simulator frame rate = $0.1 Hz$

    ## Episode End

    An episode will end if any of the following conditions are met:

    * Terminated: all points around the chief have been inspected
    * Terminated: the agent moves within a `crash_region_radius = 10` meter radius around the chief
    * Truncated: the agent exceeds a `max_distance = 800` meter radius away from the chief
    * Truncated: the maximum number of timesteps, `max_timesteps = 12236`, is reached

    The episode is considered done and successful if and only if all points have been inspected.

    ## Information

    `step()` and `reset()` return a dict with the following keys:
    * reward_components - a dict of reward component string names to their last computed float values.
    * status - a string descirbing the status of the current episode.

    Statuses for the episode are either "Running" or describe a unique terminal state. Terminal
    states can be one of the following: "Success", "Crash", "Out of Bounds", "Timeout".
    
    ## References

    <a id="1">[1]</a>
    Clohessy, W., and Wiltshire, R., “Terminal Guidance System for Satellite Rendezvous,” *Journal of the Aerospace Sciences*, Vol. 27, No. 9, 1960, pp. 653–658.

    <a id="2">[2]</a>
    Dunlap, K., Mote, M., Delsing, K., and Hobbs, K. L., “Run Time Assured Reinforcement Learning for Safe Satellite
    Docking,” *Journal of Aerospace Information Systems*, Vol. 20, No. 1, 2023, pp. 25–36. [https://doi.org/10.2514/1.I011126](https://doi.org/10.2514/1.I011126).

    <a id="3">[3]</a>
    Gaudet, B., Linares, R., and Furfaro, R., “Adaptive Guidance and Integrated Navigation with Reinforcement Meta-Learning,”
    *CoRR*, Vol. abs/1904.09865, 2019. URL [http://arxiv.org/abs/1904.09865](http://arxiv.org/abs/1904.09865).

    <a id="4">[4]</a>
    Battin, R. H., “An introduction to the mathematics and methods of astrodynamics,” 1987.

    <a id="5">[5]</a>
    Campbell, T., Furfaro, R., Linares, R., and Gaylor, D., “A Deep Learning Approach For Optical Autonomous Planetary Relative
    Terrain Navigation,” 2017.

    <a id="6">[6]</a>
    Furfaro, R., Bloise, I., Orlandelli, M., Di Lizia, P., Topputo, F., and Linares, R., “Deep Learning for Autonomous Lunar
    Landing,” 2018.

    <a id="7">[7]</a>
    Lei, H. H., Shubert, M., Damron, N., Lang, K., and Phillips, S., “Deep reinforcement Learning for Multi-agent Autonomous
    Satellite Inspection,” *AAS Guidance Navigation and Control Conference*, 2022.

    <a id="8">[8]</a>
    Aurand, J., Lei, H., Cutlip, S., Lang, K., and Phillips, S., “Exposure-Based Multi-Agent Inspection of a Tumbling Target Using
    Deep Reinforcement Learning,” *AAS Guidance Navigation and Control Conference*, 2023.

    <a id="9">[9]</a>
    Brandonisio, A., Lavagna, M., and Guzzetti, D., “Reinforcement Learning for Uncooperative Space Objects Smart Imaging
    Path-Planning,” The *Journal of the Astronautical Sciences*, Vol. 68, No. 4, 2021, pp. 1145–1169. [https://doi.org/10.1007/s40295-021-00288-7](https://doi.org/10.1007/s40295-021-00288-7).
    """

    def __init__(
        self,
        success_threshold: float = 99,
        crash_radius: float = 15,
        max_distance: float = 800,
        max_time: float = 12236,
    ) -> None:
        # Each spacecraft obs = [x, y, z, v_x, v_y, v_z, theta_sun, n, x_ups, y_ups, z_ups]
        self.observation_space = gym.spaces.Box(
            np.concatenate(
                (
                    [-np.inf] * 3,  # position
                    [-np.inf] * 3,  # velocity
                    [0],  # sun angle
                    [0],  # num inspected
                    [-1] * 3,  # nearest cluster
                )
            ),
            np.concatenate(
                (
                    [np.inf] * 3,  # position
                    [np.inf] * 3,  # velocity
                    [2 * np.pi],  # sun angle
                    [100],  # num inspected
                    [1] * 3,  # nearest cluster
                )
            ),
            shape=(11,),
        )

        self.action_space = gym.spaces.Box(
            -1, 1, shape=(3,)
        )  # only the deputy is controlled

        # Environment parameters
        self.crash_radius = crash_radius
        self.max_distance = max_distance
        self.max_time = max_time
        self.success_threshold = success_threshold

        # Episode level information
        self.prev_state = None
        self.prev_num_inspected = 0
        self.reward_components = {}
        self.status = "Running"

    def reset(
        self, *, seed: int | None = None, options: dict[str, typing.Any] | None = None
    ) -> tuple[typing.Any, dict[str, typing.Any]]:
        super().reset(seed=seed, options=options)
        self._init_sim()  # sim is light enough we just reconstruct it
        self.simulator.reset()
        obs, info = self._get_obs(), self._get_info()
        self.prev_state = None
        self.prev_num_inspected = 0
        return obs, info

    def step(
        self, action: typing.Any
    ) -> tuple[typing.Any, typing.SupportsFloat, bool, bool, dict[str, typing.Any]]:
        # Store previous simulator state
        self.prev_state = self.sim_state.copy()
        self.prev_num_inspected = (
            self.chief.inspection_points.get_num_points_inspected()
        )

        # Update simulator state
        self.deputy.add_control(action)
        self.simulator.step()

        # Get info from simulator
        observation = self._get_obs()
        reward = self._get_reward()
        terminated = self._get_terminated()
        truncated = self._get_truncated()
        info = self._get_info()
        return observation, reward, terminated, truncated, info

    def _init_sim(self):
        # Initialize spacecraft, sun, and simulator
        self.chief = sim.Target(
            name="chief",
            num_points=100,
            radius=1,
        )
        self.deputy = sim.Inspector(
            name="deputy",
            position=utils.polar_to_cartesian(
                r=self.np_random.uniform(50, 100),
                theta=self.np_random.uniform(0, 2 * np.pi),
                phi=self.np_random.uniform(-np.pi / 2, np.pi / 2),
            ),
            velocity=utils.polar_to_cartesian(
                r=self.np_random.uniform(0, 0.8),
                theta=self.np_random.uniform(0, 2 * np.pi),
                phi=self.np_random.uniform(-np.pi / 2, np.pi / 2),
            ),
            fov=np.pi,
            focal_length=1,
        )
        self.sun = sim.Sun(theta=self.np_random.uniform(0, 2 * np.pi))
        self.simulator = sim.InspectionSimulator(
            frame_rate=0.1,
            inspectors=[self.deputy],
            targets=[self.chief],
            sun=self.sun,
        )

    def _get_obs(self):
        obs = self.observation_space.sample()
        obs[:3] = self.deputy.position
        obs[3:6] = self.deputy.velocity
        obs[6] = self.sun.theta
        obs[7] = self.chief.inspection_points.get_num_points_inspected()
        obs[8:11] = self.chief.inspection_points.kmeans_find_nearest_cluster(
            camera=self.deputy.camera, sun=self.sun
        )
        return obs

    def _get_info(self):
        return {
            "reward_components": self.reward_components,
            "status": self.status
        }

    def _get_reward(self):
        reward = 0

        # Dense rewards
        points_reward = r.observed_points_reward(
            chief=self.chief, prev_num_inspected=self.prev_num_inspected
        )
        self.reward_components["observed_points"] = points_reward
        reward += points_reward

        delta_v_reward = r.delta_v_reward(
            v=self.deputy.velocity, prev_v=self.prev_state["deputy"][3:6]
        )
        self.reward_components["delta_v"] = delta_v_reward
        reward += delta_v_reward

        # Sparse rewards
        success_reward = r.inspection_success_reward(
            chief=self.chief,
            total_points=self.success_threshold,
        )
        self.reward_components["success"] = success_reward
        reward += success_reward

        crash_reward = r.crash_reward(
            chief=self.chief,
            deputy=self.deputy,
            crash_radius=self.crash_radius,
        )
        self.reward_components["crash"] = crash_reward
        reward += crash_reward

        return reward

    def _get_terminated(self):
        # Get state info
        d = utils.rel_dist(pos1=self.chief.position, pos2=self.deputy.position)

        # Determine if in terminal state
        crash = d < self.crash_radius
        all_inspected = (
            self.chief.inspection_points.get_num_points_inspected()
            >= self.success_threshold
        )

        # Update Status
        if crash:
            self.status = "Crash"
        elif all_inspected:
            self.status = "Success"

        return crash or all_inspected

    def _get_truncated(self):
        d = utils.rel_dist(pos1=self.chief.position, pos2=self.deputy.position)
        timeout = self.simulator.sim_time > self.max_time
        oob = d > self.max_distance

        # Update Status
        if oob:
            self.status = "Out of Bounds"
        elif timeout:
            self.status = "Timeout"

        return timeout or oob

    @property
    def sim_state(self) -> dict:
        state = {
            "deputy": self.deputy.state,
            "chief": self.chief.state,
        }
        return state



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements functions that define the terminal conditions
for the Docking Environment.
"""

import gymnasium
import numpy as np
from corl.dones.done_func_base import DoneFuncBase, DoneFuncBaseValidator, DoneStatusCodes
from corl.simulators.common_platform_utils import get_platform_by_name

from safe_autonomy_sims.utils import get_relative_position, get_relative_velocity, max_vel_violation


class SuccessfulDockingDoneValidator(DoneFuncBaseValidator):
    """
    A configuration validator for the SuccessfulDockingDoneFunction.

    Attributes
    ----------
    docking_region_radius : float
        The radius of the docking region in meters.
    velocity_threshold : float
        The maximum tolerated velocity within docking region without crashing.
    threshold_distance : float
        The distance at which the velocity constraint reaches a minimum (typically the docking region radius).
    slope : float
        The slope of the linear region of the velocity constraint function.
    mean_motion : float
        Orbital mean motion of Hill's reference frame's circular orbit in rad/s
    lower_bound : bool
        If True, the function enforces a minimum velocity constraint on the agent's platform.
    reference_position_sensor_name: str
        The name of the sensor responsible for returning the relative position of a reference entity.
    reference_velocity_sensor_name: str
        The name of the sensor responsible for returning the relative velocity of a reference entity.
    """

    docking_region_radius: float
    velocity_threshold: float
    threshold_distance: float
    mean_motion: float = 0.001027
    lower_bound: bool = False
    slope: float = 2.0
    reference_position_sensor_name: str = "reference_position"
    reference_velocity_sensor_name: str = "reference_velocity"


class SuccessfulDockingDoneFunction(DoneFuncBase):
    """
    A done function that determines if the deputy has successfully
    docked with the chief.

    Attributes
    ----------
    config: SuccessfulDockingDoneValidator
        The function's validated configuration parameters
    """

    def __init__(self, **kwargs) -> None:
        self.config: SuccessfulDockingDoneValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        """
        Returns configuration validator object for this done
        function.

        Returns
        -------
        SuccessfulDockingDoneValidator
            Config validator for the SuccessfulDockingDoneFunction.
        """
        return SuccessfulDockingDoneValidator

    def __call__(
        self,
        observation,
        action,
        next_observation,
        next_state,
        observation_space: gymnasium.spaces.dict.Dict,
        observation_units: gymnasium.spaces.dict.Dict,
    ) -> bool:
        """
        Parameters
        ----------
        observation : np.ndarray
            np.ndarray describing the current observation
        action : np.ndarray
            np.ndarray describing the current action
        next_observation : np.ndarray
            np.ndarray describing the incoming observation
        next_state : np.ndarray
            np.ndarray describing the incoming state
        observation_space : gymnasium.spaces.dict.Dict
            The agent observation space.
        observation_units : gymnasium.spaces.dict.Dict
            The units of the observations in the observation space. This may be None.

        Returns
        -------
        done : bool
            Dictionary containing the done condition for the current agent.
        """

        # eventually will include velocity constraint

        # Get relatative position + velocity between platform and docking region
        platform = get_platform_by_name(next_state, self.config.platform_name)
        relative_position = get_relative_position(platform, self.config.reference_position_sensor_name)
        relative_velocity = get_relative_velocity(platform, self.config.reference_velocity_sensor_name)
        distance = np.linalg.norm(relative_position)

        in_docking = distance <= self.config.docking_region_radius

        violated, _ = max_vel_violation(
            relative_position,
            relative_velocity,
            self.config.velocity_threshold,
            self.config.threshold_distance,
            self.config.mean_motion,
            self.config.lower_bound,
            slope=self.config.slope
        )

        done = bool(in_docking and not violated)

        if done:
            next_state.episode_state[self.config.platform_name][self.name] = DoneStatusCodes.WIN

        return done


class DockingVelocityLimitDoneFunctionValidator(DoneFuncBaseValidator):
    """
    A configuration validator for the DockingVelocityLimitDoneFunction.

    Attributes
    ----------
    docking_region_radius : float
        The radius of the docking region in meters.
    velocity_threshold : float
        The maximum tolerated velocity within docking region without crashing.
    threshold_distance : float
        The distance at which the velocity constraint reaches a minimum (typically the docking region radius).
    slope : float
        The slope of the linear region of the velocity constraint function.
    mean_motion : float
        Orbital mean motion of Hill's reference frame's circular orbit in rad/s
    lower_bound : bool
        If True, the function enforces a minimum velocity constraint on the agent's platform.
    reference_position_sensor_name: str
        The name of the sensor responsible for returning the relative position of a reference entity.
    reference_velocity_sensor_name: str
        The name of the sensor responsible for returning the relative velocity of a reference entity.
    """
    velocity_threshold: float
    threshold_distance: float
    mean_motion: float = 0.001027
    lower_bound: bool = False
    slope: float = 2.0
    reference_position_sensor_name: str = "reference_position"
    reference_velocity_sensor_name: str = "reference_velocity"


class DockingVelocityLimitDoneFunction(DoneFuncBase):
    """
    This done function determines whether the velocity limit
    has been exceeded.

    Attributes
    ----------
    config: DockingVelocityLimitDoneFunctionValidator
        The function's validated configuration parameters
    """

    def __init__(self, **kwargs) -> None:
        self.config: DockingVelocityLimitDoneFunctionValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        """
        Returns configuration validator object for this done
        function.

        Returns
        -------
        DockingVelocityLimitDoneFunctionValidator
            Config validator for the DockingVelocityLimitDoneFunction.
        """
        return DockingVelocityLimitDoneFunctionValidator

    def __call__(
        self,
        observation,
        action,
        next_observation,
        next_state,
        observation_space: gymnasium.spaces.dict.Dict,
        observation_units: gymnasium.spaces.dict.Dict,
    ) -> bool:
        """
        Parameters
        ----------
        observation : np.ndarray
            np.ndarray describing the current observation
        action : np.ndarray
            np.ndarray describing the current action
        next_observation : np.ndarray
            np.ndarray describing the incoming observation
        next_state : np.ndarray
            np.ndarray describing the incoming state
        observation_space : gymnasium.spaces.dict.Dict
            The agent observation space.
        observation_units : gymnasium.spaces.dict.Dict
            The units of the observations in the observation space. This may be None.

        Returns
        -------
        done : bool
            Dictionary containing the done condition for the current agent.
        """

        # Get relatative position + velocity between platform and docking region
        platform = get_platform_by_name(next_state, self.config.platform_name)
        relative_position = get_relative_position(platform, self.config.reference_position_sensor_name)
        relative_velocity = get_relative_velocity(platform, self.config.reference_velocity_sensor_name)

        violated, _ = max_vel_violation(
            relative_position,
            relative_velocity,
            self.config.velocity_threshold,
            self.config.threshold_distance,
            self.config.mean_motion,
            self.config.lower_bound,
            slope=self.config.slope,
        )

        done = bool(violated)
        if done:
            next_state.episode_state[self.config.platform_name][self.name] = DoneStatusCodes.LOSE

        return done


class DockingRelativeVelocityConstraintDoneFunctionValidator(DoneFuncBaseValidator):
    """
    A configuration validator for the DockingRelativeVelocityConstraintDoneFunction.

    Attributes
    ----------
    velocity_threshold : float
        The maximum tolerated velocity within docking region without crashing.
    threshold_distance : float
        The distance at which the velocity constraint reaches a minimum (typically the docking region radius).
    slope : float
        The slope of the linear region of the velocity constraint function.
    mean_motion : float
        Orbital mean motion of Hill's reference frame's circular orbit in rad/s
    lower_bound : bool
        If True, the function enforces a minimum velocity constraint on the agent's platform.
    reference_position_sensor_name: str
        The name of the sensor responsible for returning the relative position of a reference entity.
    reference_velocity_sensor_name: str
        The name of the sensor responsible for returning the relative velocity of a reference entity.
    """
    velocity_threshold: float
    threshold_distance: float
    mean_motion: float = 0.001027
    lower_bound: bool = False
    slope: float = 2.0
    reference_position_sensor_name: str = "reference_position"
    reference_velocity_sensor_name: str = "reference_velocity"


# needs a reference object
class DockingRelativeVelocityConstraintDoneFunction(DoneFuncBase):
    """
    A done function that checks if the docking velocity relative
    to a target object has exceeded a certain specified threshold
    velocity.

    Attributes
    ----------
    config: DockingRelativeVelocityConstraintDoneFunctionValidator
        The function's validated configuration parameters
    """

    def __init__(self, **kwargs) -> None:
        self.config: DockingRelativeVelocityConstraintDoneFunctionValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        """
        Returns configuration validator object for this done
        function.

        Returns
        -------
        DockingRelativeVelocityConstraintDoneFunctionValidator
            Configuration validator for the DockingRelativeVelocityConstraintDoneFunction
        """

        return DockingRelativeVelocityConstraintDoneFunctionValidator

    def __call__(
        self,
        observation,
        action,
        next_observation,
        next_state,
        observation_space: gymnasium.spaces.dict.Dict,
        observation_units: gymnasium.spaces.dict.Dict,
    ) -> bool:
        """
        Parameters
        ----------
        observation : np.ndarray
            np.ndarray describing the current observation
        action : np.ndarray
            np.ndarray describing the current action
        next_observation : np.ndarray
            np.ndarray describing the incoming observation
        next_state : np.ndarray
            np.ndarray describing the incoming state
        observation_space : gymnasium.spaces.dict.Dict
            The agent observation space.
        observation_units : gymnasium.spaces.dict.Dict
            The units of the observations in the observation space. This may be None.

        Returns
        -------
        done : bool
            Dictionary containing the done condition for the current agent.
        """
        # eventually will include velocity constraint

        # Get relatative position + velocity between platform and docking region
        platform = get_platform_by_name(next_state, self.config.platform_name)
        relative_position = get_relative_position(platform, self.config.reference_position_sensor_name)
        relative_velocity = get_relative_velocity(platform, self.config.reference_velocity_sensor_name)

        violated, _ = max_vel_violation(
            relative_position,
            relative_velocity,
            self.config.velocity_threshold,
            self.config.threshold_distance,
            self.config.mean_motion,
            self.config.lower_bound,
            slope=self.config.slope,
        )

        done = bool(violated)

        if done:
            next_state.episode_state[self.config.platform_name][self.name] = DoneStatusCodes.LOSE

        return done



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning Core (CoRL) Safe Autonomy Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module contains the InspectionSimulator for interacting with the CWH inspection task simulator
"""
import copy
import math
import typing

import matplotlib.pyplot as plt
import numpy as np
from corl.libraries.plugin_library import PluginLibrary
from corl.libraries.units import Quantity
from pydantic import BaseModel, validator
from ray.rllib import BaseEnv
from ray.rllib.algorithms.callbacks import DefaultCallbacks
from ray.rllib.policy import Policy
from ray.rllib.utils.typing import PolicyID
from safe_autonomy_simulation import Entity
from safe_autonomy_simulation.sims.spacecraft import CWHSpacecraft, SixDOFSpacecraft
from scipy.spatial.transform import Rotation
from sklearn.cluster import KMeans

import safe_autonomy_sims.simulators.illumination_functions as illum
from safe_autonomy_sims.platforms.cwh.cwh_platform import CWHPlatform, CWHSixDOFPlatform
from safe_autonomy_sims.simulators.saferl_simulator import (
    SafeRLSimulator,
    SafeRLSimulatorResetValidator,
    SafeRLSimulatorState,
    SafeRLSimulatorValidator,
)


class IlluminationValidator(BaseModel):
    """
    A configuration validator for illumination parameters

    Attributes
    ----------
    mean_motion: float
        A float representing the mean motion of the spacecraft in Low Earth Orbit (LEO) in [RADIANS/SECOND].
    avg_rad_Earth2Sun: float
        A float representing the average distance between the Earth and the Sun in [METERS].
    light_properties: dict
        A dict containing the ambient, specular and diffuse light properties.
    chief_properties: dict
        A dict containing the ambient, specular, diffuse, shininess and reflective properties of the chief spacecraft.
    resolution: list
        A list containing the resolution of the sensor, represented by x and y pixel density respectively.
    focal_length: float
        A float representing the focal length of the sensor in [METERS]. The virtual image is created a
        distance of focal length away from the sensor origin.
    bin_ray_flag: bool
        A bool flag for utilization of "binary ray" vs. illumination features.
    """

    mean_motion: float = 0.001027
    avg_rad_Earth2Sun: float = 150000000000
    light_properties: dict
    chief_properties: dict
    resolution: list
    focal_length: float
    bin_ray_flag: bool
    render_flag_3d: bool = False
    render_flag_subplots: bool = False
    pixel_pitch: float


class InspectionPointsValidator(BaseModel):
    """
    A configuration validator for an InspectionPoints object.

    Attributes
    ----------
    num_points: int
        The number of inspectable points maintained.
    radius: float
        The radius of the sphere on which the points will be generated.
    points_algorithm: str
        The name of the algorithm used to generate initial point positions.
    sensor_fov: float
        The field of view of the inspector's camera sensor in radians.
        Default is 2 * pi (everything is in view).
    initial_sensor_unit_vec: list
        The initial direction the inspector's camera sensor is pointing.
    illumination_params: typing.Union[IlluminationValidator, None]
        The parameters defining lighting of the environment.
    """
    num_points: int
    radius: float
    points_algorithm: str = "cmu"
    sensor_fov: float = 2 * np.pi
    initial_sensor_unit_vec: list = [1., 0., 0.]
    illumination_params: typing.Union[IlluminationValidator, None] = None

    @validator("points_algorithm")
    def valid_algorithm(cls, v):
        """
        Check if provided algorithm is a valid choice.
        """
        valid_algs = ["cmu", "fibonacci"]
        if v not in valid_algs:
            raise ValueError(f"field points_algorithm must be one of {valid_algs}")
        return v


class InspectionPoints:
    """
    A class maintaining the inspection status of an entity.
    """

    def __init__(self, parent_entity: CWHSpacecraft, priority_vector: np.ndarray, **kwargs):
        self.config: InspectionPointsValidator = self.get_validator()(**kwargs)
        self.sun_angle = 0.0
        self.clock = 0.0
        self.parent_entity = parent_entity
        self.priority_vector = priority_vector
        self.init_priority_vector = copy.deepcopy(self.priority_vector)
        (self._default_points_position_dict, self.points_position_dict, self.points_inspected_dict,
         self.points_weights_dict) = self._add_points()
        self.last_points_inspected = 0
        self.last_cluster = None

    @staticmethod
    def get_validator() -> typing.Type[InspectionPointsValidator]:
        """
        Get the validator used to validate the kwargs passed to BaseAgent.

        Returns
        -------
        BaseAgentParser
            A BaseAgent kwargs parser and validator.
        """
        return InspectionPointsValidator

    def _add_points(self):
        """
        Generate a map of inspection point coordinates to inspected state.

        Returns
        -------
        points_dict
            dict of points_dict[cartesian_point] = initial_inspected_state
        """
        if self.config.points_algorithm == "cmu":
            points_alg = InspectionPoints.points_on_sphere_cmu
        else:
            points_alg = InspectionPoints.points_on_sphere_fibonacci
        points = points_alg(self.config.num_points, self.config.radius)  # TODO: HANDLE POSITION UNITS*
        points_position_dict = {}
        points_inspected_dict = {}
        points_weights_dict = {}
        for i, point in enumerate(points):
            points_position_dict[i] = point
            points_inspected_dict[i] = False
            points_weights_dict[i] = np.arccos(
                np.dot(-self.priority_vector, point) / (np.linalg.norm(-self.priority_vector) * np.linalg.norm(point))
            ) / np.pi

        # Normalize weighting
        total_weight = sum(list(points_weights_dict.values()))
        points_weights_dict = {k: w / total_weight for k, w in points_weights_dict.items()}

        default_points_position = copy.deepcopy(points_position_dict)

        return default_points_position, points_position_dict, points_inspected_dict, points_weights_dict

    # inspected or not
    def update_points_inspection_status(self, inspector_entity):
        """
        Update the inspected state of all inspection points given an inspector's position.

        Parameters
        ----------
        inspector_entity : PyObject
            sim entity of inspector platform

        Returns
        -------
        None
        """
        # calculate h of the spherical cap (inspection zone)
        position = inspector_entity.position
        if isinstance(inspector_entity, SixDOFSpacecraft):
            rotation = Rotation.from_quat(inspector_entity.orientation)
            r_c = rotation.apply(self.config.initial_sensor_unit_vec)
        else:
            r_c = -position
        r_c = r_c / np.linalg.norm(r_c)

        r = self.config.radius
        rt = np.linalg.norm(position)
        h = 2 * r * ((rt - r) / (2 * rt))

        p_hat = position / np.linalg.norm(position)  # position unit vector (inspection zone cone axis)

        for point_id, point_position in self.points_position_dict.items():  # pylint: disable=too-many-nested-blocks
            # check that point hasn't already been inspected
            if not self.points_inspected_dict[point_id]:
                p = point_position - position
                cos_theta = np.dot(p / np.linalg.norm(p), r_c)
                angle_to_point = np.arccos(cos_theta)
                # If the point can be inspected (within FOV)
                if angle_to_point <= self.config.sensor_fov / 2:
                    # if no illumination params detected
                    if not self.config.illumination_params:
                        # project point onto inspection zone axis and check if in inspection zone
                        if np.dot(point_position, p_hat) >= r - h:
                            self.points_inspected_dict[point_id] = inspector_entity.name
                    else:
                        mag = np.dot(point_position, p_hat)
                        if mag >= r - h:
                            r_avg = self.config.illumination_params.avg_rad_Earth2Sun
                            chief_properties = self.config.illumination_params.chief_properties
                            light_properties = self.config.illumination_params.light_properties
                            current_theta = self.sun_angle
                            if self.config.illumination_params.bin_ray_flag:
                                if illum.check_illum(point_position, current_theta, r_avg, r):
                                    self.points_inspected_dict[point_id] = inspector_entity.name
                            else:
                                RGB = illum.compute_illum_pt(
                                    point_position, current_theta, position, r_avg, r, chief_properties, light_properties
                                )
                                if illum.evaluate_RGB(RGB):
                                    self.points_inspected_dict[point_id] = inspector_entity.name

    def kmeans_find_nearest_cluster(self, position):
        """Finds nearest cluster of uninspected points using kmeans clustering

        Parameters
        ----------
        position : list
            position vector

        Returns
        -------
        list
            unit vector pointing to nearest cluster
        """
        uninspected = []
        for point_id, inspected in self.points_inspected_dict.items():
            point_position = self.points_position_dict[point_id]
            if not inspected:
                if self.config.illumination_params:
                    if self.check_if_illuminated(point_position, position):
                        uninspected.append(point_position)
                else:
                    uninspected.append(point_position)
        if len(uninspected) == 0:
            out = np.array([0., 0., 0.])
        else:
            n = math.ceil(len(uninspected) / 10)
            data = np.array(uninspected)
            if self.last_cluster is None:
                init = np.zeros((n, 3))
            else:
                if n > self.last_cluster.shape[0]:
                    idxs = np.random.choice(self.last_cluster.shape[0], size=n - self.last_cluster.shape[0])
                    new = np.array(uninspected)[idxs, :]
                    init = np.vstack((self.last_cluster, new))
                else:
                    init = self.last_cluster[0:n, :]
            kmeans = KMeans(
                init=init,
                n_clusters=n,
                n_init=10,
                max_iter=50,
            )
            kmeans.fit(data)
            self.last_cluster = kmeans.cluster_centers_
            dist = []
            for center in self.last_cluster:
                dist.append(np.linalg.norm(position - center))
            out = kmeans.cluster_centers_[np.argmin(dist)]
            out = out / np.linalg.norm(out)
        return out

    def check_if_illuminated(self, point, position):
        """Check if point is illuminated

        Parameters
        ----------
        point : list
            point position vector
        position : list
            agent position vector

        Returns
        -------
        bool
            True if point is illuminated, False otherwise
        """
        r = self.config.radius
        r_avg = self.config.illumination_params.avg_rad_Earth2Sun
        chief_properties = self.config.illumination_params.chief_properties
        light_properties = self.config.illumination_params.light_properties
        current_theta = self.sun_angle
        if self.config.illumination_params.bin_ray_flag:
            illuminated = illum.check_illum(point, current_theta, r_avg, r)
        else:
            RGB = illum.compute_illum_pt(point, current_theta, position, r_avg, r, chief_properties, light_properties)
            illuminated = illum.evaluate_RGB(RGB)
        return illuminated

    @staticmethod
    def points_on_sphere_fibonacci(num_points: int, radius: float) -> list:
        """
        Generate a set of equidistant points on sphere using the
        Fibonacci Sphere algorithm: https://arxiv.org/pdf/0912.4540.pdf

        Parameters
        ----------
        num_points: int
            number of points to attempt to place on a sphere
        radius: float
            radius of the sphere

        Returns
        -------
        points: list
            Set of equidistant points on sphere in cartesian coordinates
        """
        points = []
        phi = math.pi * (3. - math.sqrt(5.))  # golden angle in radians

        for i in range(num_points):
            y = 1 - (i / float(num_points - 1)) * 2  # y goes from 1 to -1
            r = math.sqrt(1 - y * y)  # radius at y

            theta = phi * i  # golden angle increment

            x = math.cos(theta) * r
            z = math.sin(theta) * r

            points.append(radius * np.array([x, y, z]))

        return points

    @staticmethod
    def points_on_sphere_cmu(num_points: int, radius: float) -> list:
        """
        Generate a set of equidistant points on a sphere using the algorithm
        in https://www.cmu.edu/biolphys/deserno/pdf/sphere_equi.pdf. Number
        of points may not be exact.

        Mostly the same as CMU algorithm, most important tweak is that the constant "a" should not depend on r
        (Paper assumed r = 1)

        Parameters
        ----------
        num_points: int
            number of points to attempt to place on a sphere
        radius: float
            radius of the sphere

        Returns
        -------
        points: list
            Set of equidistant points on sphere in cartesian coordinates
        """
        points = []

        a = 4.0 * math.pi * (1 / num_points)
        d = math.sqrt(a)
        m_theta = int(round(math.pi / d))
        d_theta = math.pi / m_theta
        d_phi = a / d_theta

        for m in range(0, m_theta):
            theta = math.pi * (m + 0.5) / m_theta
            m_phi = int(round(2.0 * math.pi * math.sin(theta) / d_phi))
            for n in range(0, m_phi):
                phi = 2.0 * math.pi * n / m_phi

                x = radius * math.sin(theta) * math.cos(phi)
                y = radius * math.sin(theta) * math.sin(phi)
                z = radius * math.cos(theta)

                points.append(np.array([x, y, z]))

        return points

    def update_points_position(self):
        """
        Update the locations of the points on the chief after rotation
        """

        # get parent entity info
        parent_position = self.parent_entity.position
        parent_orientation = self.parent_entity.orientation
        parent_orientation = Rotation.from_quat(parent_orientation)

        self.priority_vector = parent_orientation.apply(self.init_priority_vector)

        for point_id, default_position in self._default_points_position_dict.items():
            # rotate about origin
            new_position = parent_orientation.apply(default_position)
            # translate from origin
            new_position = new_position + parent_position
            self.points_position_dict[point_id] = new_position

    # getters / setters
    def get_num_points_inspected(self, inspector_entity: Entity = None):
        """Get total number of points inspected"""
        num_points = 0
        if inspector_entity:
            # count number of points inspected by the provided entity
            for _, point_inspector_entity in self.points_inspected_dict.items():
                num_points += 1 if point_inspector_entity == inspector_entity.name else 0
        else:
            # count the total number of points inspected
            for _, point_inspector_entity in self.points_inspected_dict.items():
                num_points += 1 if point_inspector_entity else 0

        return num_points

    def get_percentage_of_points_inspected(self, inspector_entity: Entity = None):
        """Get the percentage of points inspected"""
        total_num_points = len(self.points_inspected_dict.keys())

        if inspector_entity:
            percent = self.get_num_points_inspected(inspector_entity=inspector_entity) / total_num_points
        else:
            percent = self.get_num_points_inspected() / total_num_points
        return percent

    def get_cluster_location(self, inspector_position):
        """Get the location of the nearest cluster of uninspected points"""
        return self._kmeans_find_nearest(inspector_position)

    def get_total_weight_inspected(self, inspector_entity: Entity = None):
        """Get total weight of points inspected"""
        weights = 0
        if inspector_entity:
            for point_inspector_entity, weight in zip(self.points_inspected_dict.values(), self.points_weights_dict.values()):
                weights += weight if point_inspector_entity == inspector_entity.name else 0.
        else:
            for point_inspector_entity, weight in zip(self.points_inspected_dict.values(), self.points_weights_dict.values()):
                weights += weight if point_inspector_entity else 0.
        return weights

    def set_sun_angle(self, sun_angle: np.ndarray):
        """Get the current sun angle"""
        self.sun_angle = float(sun_angle)


class InspectionSimulatorState(SafeRLSimulatorState):
    """
    The basemodel for the state of the InspectionSimulator.

    Attributes
    ----------
    inspection_points_map: dict
        The dictionary containing the points the agent needs to inspect.
        Keys: point_id int. Values: InspectionPoints objects.
    total_steps: int
        The total number of steps simulated since the simulator was initialized
    delta_v_scale: float
        The scale of the delta_v reward. This value is updated over time, based on the value of inspected_points_value.
        See safe_autonomy_sims/rewards/cwh/inspection_rewards for more details.
        Note that this feature is experimental, and currently does not synchronize between workers.
    sun_angle: typing.Union[Quantity, float]
        Angle of the Sun in the x-y plane
    priority_vector: np.ndarray
        Vector indicating priority of points to be inspected
    """
    inspection_points_map: typing.Dict[str, InspectionPoints]
    total_steps: int
    delta_v_scale: float
    sun_angle: typing.Union[Quantity, float] = 0.0
    priority_vector: np.ndarray

    class Config:
        """Allow arbitrary types for Parameter"""
        arbitrary_types_allowed = True


class InspectionSimulatorValidator(SafeRLSimulatorValidator):
    """
    A configuration validator for the InspectionSimulator config.

    Attributes
    ----------
    illumination_params: IlluminationValidator
        dict of illumination parameters, described above
    steps_until_update: int
        Number of steps between updates to the delta_v_scale
    delta_v_updater_criteria: str
        Criteria for updating the delta-v scale.
        Either 'score' for total points score, or 'count' for total number of points
    delta_v_scale_bounds: list
        lower and upper bounds for the value of delta_v_scale
    delta_v_scale_step: float
        The amount to advance/retract the delta_v_scale by at each update
    inspected_points_update_bounds: list
        bounds for when to update the delta_v_scale.
        if inspected_points_value >= inspected_points_update_bounds[1], delta_v_scale is advanced by delta_v_scale_step
        if inspected_points_value <= inspected_points_update_bounds[0], delta_v_scale is retracted by delta_v_scale_step
    sensor_fov: float
        field of view of the sensor (radians).
        Default is 2 * pi (everything is in view).
    initial_sensor_unit_vec: list
        If using the 6DOF spacecraft model, initial unit vector along sensor boresight.
        By default [1., 0., 0.]
    inspection_points_map: dict
        A map of entity name strings to InspectionPoints objects, which track the inspection progress of each entity.
    """

    illumination_params: typing.Union[IlluminationValidator, None] = None
    steps_until_update: int
    delta_v_updater_criteria: str = 'count'
    delta_v_scale_bounds: list
    delta_v_scale_step: float
    inspected_points_update_bounds: list
    sensor_fov: float = 2 * np.pi
    initial_sensor_unit_vec: list = [1., 0., 0.]
    inspection_points_map: typing.Dict[str, InspectionPointsValidator]

    class Config:
        """Allow arbitrary types for Parameter"""
        arbitrary_types_allowed = True


class InspectionSimulatorResetValidator(SafeRLSimulatorResetValidator):
    """
    A configuration validator for the InspectionSimulator reset method.

    priority_vector_azimuth_angle: Quantity, float
        Azimuth angle of the priority vector for weighting points
    priority_vector_elevation_angle: Quantity, float
        Elevation angle of the priority vector for weighting points
    """
    priority_vector_azimuth_angle: typing.Union[Quantity, float] = 0.0
    priority_vector_elevation_angle: typing.Union[Quantity, float] = 0.0

    class Config:
        """Allow arbitrary types for Parameter"""
        arbitrary_types_allowed = True


class InspectionSimulator(SafeRLSimulator):
    """
    Simulator for CWH Inspection Task. Interfaces CWH platforms with underlying CWH entities in inspection simulation.
    """

    @property
    def get_simulator_validator(self):
        return InspectionSimulatorValidator

    @property
    def get_reset_validator(self) -> typing.Type[InspectionSimulatorResetValidator]:
        return InspectionSimulatorResetValidator

    def _construct_platform_map(self) -> dict:
        return {
            'default': (CWHSpacecraft, CWHPlatform),
            'cwh': (CWHSpacecraft, CWHPlatform),
            'six-dof': (SixDOFSpacecraft, CWHSixDOFPlatform),
        }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.inspection_points_map = {}

        if self.config.illumination_params is not None:
            if self.config.illumination_params.render_flag_3d:
                self.fig = plt.figure(1)
            if self.config.illumination_params.render_flag_subplots:
                self.fig = plt.figure(1)
                self.ax_3d = self.fig.add_subplot(2, 2, 1, projection='3d')
                self.ax_xy = self.fig.add_subplot(2, 2, 2)
                self.ax_xz = self.fig.add_subplot(2, 2, 3)
                self.ax_yz = self.fig.add_subplot(2, 2, 4)

        self.total_steps = 0
        self.inspected_points_value_list = []
        self.update_step_current = 0
        self.delta_v_scale = 0
        self.delta_v_updater_criteria = self.config.delta_v_updater_criteria
        self.steps_until_update = self.config.steps_until_update
        self.delta_v_scale_bounds = self.config.delta_v_scale_bounds
        self.delta_v_scale_step = self.config.delta_v_scale_step
        self.inspected_points_update_bounds = self.config.inspected_points_update_bounds
        self.sun_angle = 0.
        self.priority_vector = np.zeros(3)

    def create_inspection_points_map(self):
        """
        Create map of inspection points for each entity
        """
        points_map = {}
        for entity_name, inspection_points_validator in self.config.inspection_points_map.items():
            # TODO: there must be a better way to use validators
            parent_entity = self.sim_entities[entity_name]
            points_map[entity_name] = InspectionPoints(
                parent_entity=parent_entity,
                radius=inspection_points_validator.radius,
                num_points=inspection_points_validator.num_points,
                points_algorithm=inspection_points_validator.points_algorithm,
                sensor_fov=self.config.sensor_fov,
                initial_sensor_unit_vec=self.config.initial_sensor_unit_vec,
                illumination_params=self.config.illumination_params,
                priority_vector=self.priority_vector,
            )

        return points_map

    def _construct_simulator_state(self) -> dict:
        return InspectionSimulatorState(
            sim_platforms=self.sim_platforms,
            inspection_points_map=self.inspection_points_map,
            sim_time=self.clock,
            sim_entities=self.sim_entities,
            total_steps=self.total_steps,
            delta_v_scale=self.delta_v_scale,
            sun_angle=self.sun_angle,
            priority_vector=self.priority_vector,
        )

    def reset(self, config):
        super().reset(config)
        if self.config.illumination_params is not None:
            self.sun_angle = self.sim_entities['sun'].theta

        self._get_initial_priority_vector(config)

        if self.inspection_points_map:
            # calculate delta_v_scale
            # TODO: assumes first agent
            agent = list(self.inspection_points_map.keys())[0]
            final_num_points_inspected = self.inspection_points_map[agent].get_num_points_inspected()
            if final_num_points_inspected != 0:
                if self.delta_v_updater_criteria == 'score':
                    self.inspected_points_value_list.append(self.inspection_points_map[agent].get_total_weight_inspected())
                elif self.delta_v_updater_criteria == 'count':
                    self.inspected_points_value_list.append(self.inspection_points_map[agent].get_percentage_of_points_inspected())
                else:
                    raise ValueError('delta_v_updater_criteria must be either "score" or "count"')
            if self.total_steps >= self.update_step_current:
                self.update_step_current += self.steps_until_update
                mean_inspected_points_value = np.mean(self.inspected_points_value_list)
                self.inspected_points_value_list = []
                if mean_inspected_points_value >= self.inspected_points_update_bounds[1]:
                    self.delta_v_scale += self.delta_v_scale_step
                if mean_inspected_points_value <= self.inspected_points_update_bounds[0]:
                    self.delta_v_scale -= self.delta_v_scale_step

        self.delta_v_scale = np.clip(self.delta_v_scale, self.delta_v_scale_bounds[0], self.delta_v_scale_bounds[1])

        # reset points map
        self.inspection_points_map = self.create_inspection_points_map()
        for points in self.inspection_points_map.values():
            points.update_points_position()
            # TODO: assign priority vector to specific agent
            self.priority_vector = points.priority_vector
        # illuminate
        if self.config.illumination_params:
            # pass sun_angle to InspectionPoints objs
            for points in self.inspection_points_map.values():
                points.set_sun_angle(self.sun_angle)

        self._update_inspection_points_statuses()

        self._state = self._construct_simulator_state()

        self.update_sensor_measurements()
        return self._state

    def _get_initial_priority_vector(self, config):
        """Get the initial priority vector for weighting points"""
        azi = config["priority_vector_azimuth_angle"]
        if isinstance(azi, Quantity):
            azi = azi.m
        ele = config["priority_vector_elevation_angle"]
        if isinstance(ele, Quantity):
            ele = ele.m

        self.priority_vector[0] = np.cos(azi) * np.cos(ele)
        self.priority_vector[1] = np.sin(azi) * np.cos(ele)
        self.priority_vector[2] = np.sin(ele)

    def _step_update_sim_statuses(self, step_size: float):
        self.total_steps += 1
        self._state.total_steps = self.total_steps

        # update inspection points positions
        for points in self.inspection_points_map.values():
            points.update_points_position()
            # TODO: assign priority vector to specific agent
            self.priority_vector = points.priority_vector
            self._state.priority_vector = points.priority_vector

        # illuminate
        if self.config.illumination_params:
            self.sun_angle = self.sim_entities['sun'].theta
            # pass sun_angle to InspectionPoints objs
            for points in self.inspection_points_map.values():
                points.set_sun_angle(self.sun_angle)
            self._state.sun_angle = float(self.sun_angle)

        # update inspection points statuses
        self._update_inspection_points_statuses()

        if self.config.illumination_params:
            # render scene every m simulation seconds
            if (self.config.illumination_params.render_flag_3d or self.config.illumination_params.render_flag_subplots):
                current_time = self.clock
                sun_position = illum.get_sun_position(
                    current_time,
                    self.config.illumination_params.mean_motion,
                    self.sun_angle,
                    self.config.illumination_params.avg_rad_Earth2Sun
                )
                m = 10
                if (current_time % (m)) == 0:
                    if self.config.illumination_params.render_flag_3d:
                        # TODO: remove "chief" assumption
                        illum.render_3d(
                            self.fig,
                            self.agent_sim_entities["blue0"].position,
                            sun_position,
                            self.inspection_points_map["chief"].config.radius,
                            current_time,
                            m
                        )
                    if self.config.illumination_params.render_flag_subplots:
                        axes = [self.ax_3d, self.ax_xy, self.ax_xz, self.ax_yz]
                        illum.render_subplots(
                            self.fig,
                            axes,
                            self.agent_sim_entities["blue0"].position,
                            sun_position,
                            self.inspection_points_map["chief"].config.radius,
                            current_time,
                            m
                        )

    def _update_inspection_points_statuses(self):
        for inspection_entity_name, points in self.inspection_points_map.items():
            for inspector_entity_name, inspector_entity in self.agent_sim_entities.items():
                if inspection_entity_name != inspector_entity_name:
                    points.update_points_inspection_status(inspector_entity)


PluginLibrary.AddClassToGroup(InspectionSimulator, "InspectionSimulator", {})


class InspectionCallbacks(DefaultCallbacks):
    """
    Custom callbacks for the Inspection Simulator
    Log value for the delta_v_scale
    """

    def on_episode_end(self, *, worker, base_env: BaseEnv, policies: typing.Dict[PolicyID, Policy], episode, **kwargs) -> None:
        super().on_episode_end(worker=worker, base_env=base_env, policies=policies, episode=episode, **kwargs)
        env = base_env.get_sub_environments()[episode.env_id]
        episode.custom_metrics['info/delta_v_scale'] = env.simulator.delta_v_scale
        episode.custom_metrics['info/total_steps'] = env.simulator.total_steps



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This package implements the reward functions used across the
safe autonomy environments.

Rewards are used to incentivize certain [agent](../agents/index.md)
behavior during training. Over the course of a training episode,
the following reward functions take in current and previous
observations and actions and return a `RewardDict`,
allocating an appropriate reward (or penalty) to an agent based on
its behavior and the [simulation](../simulators/index.md) state.
"""



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This package implements various CoRL platforms for use with the docking
and inspection environments.

Platforms represent objects within the environment
(vehicles, robots, etc.). Platforms can contain multiple custom
`PlatformPart`s which define platform behavior. The most common
types of parts are `Controller`s and `Sensor`s. `Controller`s
apply controls to thrusters and control surfaces to interact with
environment dynamics. `Sensor`s measure various aspects of the
environment. Platform parts are often analogous to the parts used
in a physical system.
"""



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning Core (CoRL) Safe Autonomy Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements reward functions based on done status
"""
import typing
from collections import OrderedDict
from functools import partial

from corl.dones.done_func_base import DoneStatusCodes
from corl.libraries.state_dict import StateDict
from corl.rewards.episode_done import EpisodeDoneStateReward, EpisodeDoneStateRewardValidator
from pydantic import validator


class WinLoseDoneRewardValidator(EpisodeDoneStateRewardValidator):
    """A configuration validator for WinLoseDoneReward

    Attributes
    ----------
    done_name : str
        name of done to generate reward from
    done_status : str
        done status code to generate reward from
    scale : float
        reward scaling value
    """
    done_name: str
    done_status: str = "lose"
    scale: float = -1.0

    @validator('done_status')
    def status_is_win_or_lose(cls, v):
        """Validate done_status"""
        if v not in ['win', 'lose']:
            raise ValueError('Done status must be "win" or "lose"')
        return v


class WinLoseDoneReward(EpisodeDoneStateReward):
    """Reward that applies to a specific Done.  Can be applied when done status
    is "win" or "lose"
    """

    def __init__(self, **kwargs) -> None:
        self.config: WinLoseDoneRewardValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> typing.Type[WinLoseDoneRewardValidator]:
        return WinLoseDoneRewardValidator

    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: StateDict,
        next_state: StateDict,
        observation_space,
        observation_units
    ) -> float:

        reward = 0.0

        done_state = next_state.agent_episode_state.get(self.config.agent_name, {})
        for done_name, done_code in done_state.items():
            if done_name in self._already_recorded:
                continue
            self._already_recorded.add(done_name)
            self._status_codes[done_code].append(done_name)

        # this will loop starting from win and go down
        consolidate_break = False
        for done_status in DoneStatusCodes:
            for done_name in self._status_codes[done_status]:
                reward += self.get_scaling_method(
                    observation, action, next_observation, state, next_state, observation_space, observation_units, done_name, done_status
                )(self._counter)
                if self.config.consolidate:
                    consolidate_break = True
                    break
            if consolidate_break:
                break

        self._counter += 1 / next_state.sim_update_rate_hz

        return reward

    def get_scaling_method(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: StateDict,
        next_state: StateDict,
        observation_space,
        observation_units,
        done_name,
        done_code
    ) -> typing.Callable[[int], float]:
        # pylint: disable=unused-argument
        # pylint: disable=arguments-differ
        if done_name == self.config.done_name:
            if done_code.name.lower() == self.config.done_status:
                scale = self.config.scale
                return partial(self.constant_scaling, scale=scale)
            return self._status_code_func[done_code.name.lower()]

        return partial(self.constant_scaling, scale=0)



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements Run Time Assurance for Clohessy-Wiltshire spacecraft.
"""

from run_time_assurance.controller import RTABackupController
from run_time_assurance.zoo.cwh.docking_3d import (
    V0_DEFAULT,
    V1_COEF_DEFAULT,
    X_VEL_LIMIT_DEFAULT,
    Y_VEL_LIMIT_DEFAULT,
    Z_VEL_LIMIT_DEFAULT,
    Docking3dExplicitOptimizationRTA,
    Docking3dExplicitSwitchingRTA,
    Docking3dImplicitOptimizationRTA,
    Docking3dImplicitSwitchingRTA,
)
from safe_autonomy_simulation.sims.spacecraft.defaults import M_DEFAULT, N_DEFAULT

from safe_autonomy_sims.glues.rta_glue import RTAGlue, RTAGlueValidator, RTASingleton


class CWHDocking3dRTAGlueValidator(RTAGlueValidator):
    """
    A configuration validator for cwh docking 3d rta glues.

    Attributes
    ----------
    m : float, optional
        Mass in kg of spacecraft, by default M_DEFAULT.
    n : float, optional
        Orbital mean motion in rad/s of current Hill's reference frame, by default N_DEFAULT.
    v0 : float, optional
        Maximum safe docking velocity in m/s, by default V0_DEFAULT.
        v0 of v_limit = v0 + v1*n*||r||
    v1_coef : float, optional
        Coefficient of linear component of the distance dependent speed limit in 1/seconds, by default V1_COEF_DEFAULT.
        v1_coef of v_limit = v0 + v1_coef*n*||r||
    x_vel_limit : float, optional
        Max velocity magnitude in the x direction, by default X_VEL_LIMIT_DEFAULT.
    y_vel_limit : float, optional
        Max velocity magnitude in the y direction, by default Y_VEL_LIMIT_DEFAULT.
    z_vel_limit : float, optional
        Max velocity magnitude in the z direction, by default Z_VEL_LIMIT_DEFAULT.
    control_bounds_high : Union[float, np.ndarray], optional
        Upper bound of allowable control. Pass a list for element specific limit. By default, 1.
    control_bounds_low : Union[float, np.ndarray], optional
        Lower bound of allowable control. Pass a list for element specific limit. By default -1.
    """
    m: float = M_DEFAULT
    n: float = N_DEFAULT
    v0: float = V0_DEFAULT
    v1_coef: float = V1_COEF_DEFAULT
    x_vel_limit: float = X_VEL_LIMIT_DEFAULT
    y_vel_limit: float = Y_VEL_LIMIT_DEFAULT
    z_vel_limit: float = Z_VEL_LIMIT_DEFAULT
    control_bounds_high: float = 1
    control_bounds_low: float = -1


class RTAGlueCWHDocking3d(RTAGlue):
    """
    General RTA Glue to wrap CWH Docking 3d RTA from the run-time-assurance package.
    """

    def __init__(self, **kwargs):
        self.config: CWHDocking3dRTAGlueValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        return CWHDocking3dRTAGlueValidator

    def _get_rta_args(self) -> dict:
        return {
            "m": self.config.m,
            "n": self.config.n,
            "v0": self.config.v0,
            "v1_coef": self.config.v1_coef,
            "x_vel_limit": self.config.x_vel_limit,
            "y_vel_limit": self.config.y_vel_limit,
            "z_vel_limit": self.config.z_vel_limit,
            "control_bounds_high": self.config.control_bounds_high,
            "control_bounds_low": self.config.control_bounds_low,
        }


class CWHDocking3dExplicitSwitchingRTAGlueValidator(CWHDocking3dRTAGlueValidator):
    """
    A configuration validator for CWH Docking 3d Explicit Switching RTA Glue.

    Attributes
    ----------
    backup_controller : RTABackupController, optional
        Backup controller object utilized by rta module to generate backup control.
        By default, Docking2dStopLQRBackupController.
    """
    # backup_controller: RTABackupController = None
    test: int = 1


class RTAGlueCHWDocking3dExplicitSwitching(RTAGlueCWHDocking3d):
    """
    RTA Glue to wrap CWH Docking 3d Explicit Switching RTA from the run-time-assurance package.
    """

    def __init__(self, **kwargs):
        self.config: CWHDocking3dExplicitSwitchingRTAGlueValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        return CWHDocking3dExplicitSwitchingRTAGlueValidator

    def _get_singleton(self):
        return CHWDocking3dExplicitSwitchingSingleton

    def _get_rta_args(self) -> dict:
        parent_args = super()._get_rta_args()
        return {
            **parent_args,
            'backup_controller': self.config.backup_controller,
        }


class CHWDocking3dExplicitSwitchingSingleton(RTASingleton):
    """Explicit Switching Singleton"""

    def _create_rta_module(self, **rta_args):
        return Docking3dExplicitSwitchingRTA(**rta_args)


class CWHDocking3dImplicitSwitchingRTAGlueValidator(CWHDocking3dRTAGlueValidator):
    """
    A configuration validator for CWH Docking 3d Implicit Switching RTA Glue.

    Attributes
    ----------
    backup_window : float
        Duration of time in seconds to evaluate backup controller trajectory.
    backup_controller : RTABackupController, optional
        Backup controller object utilized by rta module to generate backup control.
        By default, Docking2dStopLQRBackupController.
    """
    backup_window: float = 5
    backup_controller: RTABackupController = None


class RTAGlueCHWDocking3dImplicitSwitching(RTAGlueCWHDocking3d):
    """
    RTA Glue to wrap CWH Docking 3d Implicit Switching RTA from the run-time-assurance package.
    """

    def __init__(self, **kwargs):
        self.config: CWHDocking3dImplicitSwitchingRTAGlueValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        return CWHDocking3dImplicitSwitchingRTAGlueValidator

    def _get_singleton(self):
        return CHWDocking3dImplicitSwitchingSingleton

    def _get_rta_args(self) -> dict:
        parent_args = super()._get_rta_args()
        return {
            **parent_args,
            'backup_window': self.config.backup_window,
            'backup_controller': self.config.backup_controller,
        }


class CHWDocking3dImplicitSwitchingSingleton(RTASingleton):
    """Implicit Switching Singleton"""

    def _create_rta_module(self, **rta_args):
        return Docking3dImplicitSwitchingRTA(**rta_args)


class RTAGlueCHWDocking3dExplicitOptimization(RTAGlueCWHDocking3d):
    """
    RTA Glue to wrap CWH Docking 3d Explicit Optimization RTA from the run-time-assurance package.
    """

    def _get_singleton(self):
        return CHWDocking3dExplicitOptimizationSingleton


class CHWDocking3dExplicitOptimizationSingleton(RTASingleton):
    """Explicit Optimization Singleton"""

    def _create_rta_module(self, **rta_args):
        return Docking3dExplicitOptimizationRTA(**rta_args)


class CWHDocking3dImplicitOptimizationRTAGlueValidator(CWHDocking3dRTAGlueValidator):
    """
    A configuration validator for CWH Docking 3d Implicit Optimization RTA Glue.

    Attributes
    ----------
    backup_window : float
        Duration of time in seconds to evaluate backup controller trajectory.
    num_check_all : int
        Number of points at beginning of backup trajectory to check at every sequential simulation timestep.
        Should be <= backup_window.
        Defaults to 0 as skip_length defaults to 1 resulting in all backup trajectory points being checked.
    skip_length : int
        After num_check_all points in the backup trajectory are checked, the remainder of the backup window is filled by
        skipping every skip_length points to reduce the number of backup trajectory constraints. Will always check the
        last point in the backup trajectory as well.
        Defaults to 1, resulting in no skipping.
    backup_controller : RTABackupController, optional
        Backup controller object utilized by rta module to generate backup control.
        By default, Docking2dStopLQRBackupController
    """
    backup_window: float = 5
    num_check_all: int = 5
    skip_length: int = 1
    backup_controller: RTABackupController = None


class RTAGlueCHWDocking3dImplicitOptimization(RTAGlueCWHDocking3d):
    """
    RTA Glue to wrap CWH Docking 3d Implicit Optimization RTA from the run-time-assurance package.
    """

    def __init__(self, **kwargs):
        self.config: CWHDocking3dImplicitOptimizationRTAGlueValidator
        super().__init__(**kwargs)

    @staticmethod
    def get_validator():
        return CWHDocking3dImplicitOptimizationRTAGlueValidator

    def _get_singleton(self):
        return CHWDocking3dImplicitOptimizationSingleton

    def _get_rta_args(self) -> dict:
        parent_args = super()._get_rta_args()
        return {
            **parent_args,
            'backup_window': self.config.backup_window,
            'num_check_all': self.config.num_check_all,
            'skip_length': self.config.skip_length,
            'backup_controller': self.config.backup_controller,
        }


class CHWDocking3dImplicitOptimizationSingleton(RTASingleton):
    """Implicit Optimization Singleton"""

    def _create_rta_module(self, **rta_args):
        return Docking3dImplicitOptimizationRTA(**rta_args)



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This module implements rotation sensor for use with the CWHSixDofPlatform
"""
import numpy as np
from corl.libraries.plugin_library import PluginLibrary
from corl.libraries.units import corl_get_ureg
from scipy.spatial.transform import Rotation as R

import safe_autonomy_sims.platforms.cwh.cwh_properties as cwh_props
from safe_autonomy_sims.platforms.cwh.cwh_available_platforms import CWHAvailablePlatformTypes
from safe_autonomy_sims.platforms.cwh.cwh_sensors import CWHSensor
from safe_autonomy_sims.simulators.cwh_simulator import CWHSimulator
from safe_autonomy_sims.simulators.inspection_simulator import InspectionSimulator


class QuaternionSensor(CWHSensor):
    """
    Implementation of a sensor designed to give the quaternion at any time.
    """

    def __init__(self, parent_platform, config, property_class=cwh_props.QuaternionProp):
        super().__init__(property_class=property_class, parent_platform=parent_platform, config=config)

    def _calculate_measurement(self, state):
        """
        Calculate the measurement - quaternion.

        Returns
        -------
        list of floats
            quaternion of spacecraft.
        """
        quaternion = np.array(self.parent_platform.quaternion, dtype=np.float32)
        return corl_get_ureg().Quantity(quaternion, "dimensionless")


class AngularVelocitySensor(CWHSensor):
    """
    Implementation of a sensor designed to give the angular velocity at any time.
    """

    def __init__(self, parent_platform, config, property_class=cwh_props.AngularVelocityProp):
        super().__init__(property_class=property_class, parent_platform=parent_platform, config=config)

    def _calculate_measurement(self, state):
        """
        Calculate the measurement - angular_velocity.

        Returns
        -------
        list of floats
            Angular velocity of spacecraft.
        """
        angular_velocity = np.array(self.parent_platform.angular_velocity, dtype=np.float32)
        return corl_get_ureg().Quantity(angular_velocity, "radian / second")


class OrientationUnitVectorSensor(CWHSensor):
    """
    Implementation of a sensor designed to give the canonical (or reference)
    sensor orientation unit vector.
    """

    def __init__(self, parent_platform, config, property_class=cwh_props.OrientationVectorProp):
        super().__init__(property_class=property_class, parent_platform=parent_platform, config=config)

    def _calculate_measurement(self, state):
        """
        Calculate the measurement - unit vector.

        Returns
        -------
        list of floats
            elements of unit vector that points in direction of spacecraft
            sensor.
        """
        initial_orientation = None
        try:
            initial_orientation = state.inspection_points_map['chief'].config.initial_sensor_unit_vec
        except KeyError:
            initial_orientation = np.array([0.0, 0.0, 0.0])
        if initial_orientation is None:
            initial_orientation = np.array([0.0, 0.0, 0.0])

        initial_orientation = corl_get_ureg().Quantity(np.array(initial_orientation, dtype=np.float32), "dimensionless")
        return initial_orientation


class RotatedAxesSensor(CWHSensor):
    """
    Implementation of a sensor designed to give the vectors that result from
    rotating the initial coordinate frame's unit vectors into the deputy's
    frame.

    Note: Because the canonical initial_sensor_unit_ec is [1, 0, 0] (we assume
    this hasn't been changed), we omit this unit vector.  Thus, the sensor
    returns 6 values, the x,y,x components of the remaining two unit vectors
    after rotation.
    """

    def __init__(self, parent_platform, config, property_class=cwh_props.RotatedAxesProp):
        super().__init__(property_class=property_class, parent_platform=parent_platform, config=config)

    def _calculate_measurement(self, state):
        """
        Calculate the measurement - unit vector.

        Returns
        -------
        list of floats
            elements of unit vector that points in direction of spacecraft
            sensor.
        """
        quaternion = self.parent_platform.quaternion
        r = R.from_quat(quaternion)
        v1 = np.array([0.0, 1.0, 0.0])
        v2 = np.array([0.0, 0.0, 1.0])

        rot_v1 = r.apply(v1)
        rot_v2 = r.apply(v2)

        out = np.concatenate([rot_v1, rot_v2])
        out = corl_get_ureg().Quantity(np.array(out, dtype=np.float32), "dimensionless")
        return out


for sim in [CWHSimulator, InspectionSimulator]:
    for sensor, sensor_name in zip([QuaternionSensor, AngularVelocitySensor, OrientationUnitVectorSensor, RotatedAxesSensor],
                                   ["Sensor_Quaternion", "Sensor_AngularVelocity", "Sensor_OrientationUnitVector", "Sensor_RotatedAxes"]):
        PluginLibrary.AddClassToGroup(sensor, sensor_name, {"simulator": sim, "platform_type": CWHAvailablePlatformTypes})



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

import abc
import os
import sys
import typing

import imageio.v2 as imageio
import matplotlib.pyplot as plt
import pandas as pd
from corl.evaluation.runners.section_factories.plugins.platform_serializer import PlatformSerializer
from numpy.typing import ArrayLike

from safe_autonomy_sims.evaluation.evaluation_api import run_one_evaluation


class BaseAnimationModule(abc.ABC):
    """
    Base module for making animations/plots from a checkpoint

    Parameters
    ----------
    checkpoint_path: str
        The absolute path to the checkpoint from which the policy under evaluation will be loaded
    expr_config_path: str
        The absolute path to the experiment config used in training
    task_config_path: str
        The absolute path to the task_config used in training
    platform_serializer_class: PlatformSerializer
        The PlatformSerializer subclass capable of saving the state of the Platforms used in the evaluation environment
    save_dir: str
        The absolute path to the directory in which evaluation episode(s) data will be saved
    launch_dir_of_experiment: str
        Paths in the experiment config are relative to the directory the corl.train_rl module is intended to be launched from.
        This string captures the path to the intended launch location of the experiment under evaluation.
    tmp_dir: str
        Temporary directory to save animation images to
    start_index: int
        Index to start plotting loop
    """

    def __init__(
        self,
        checkpoint_path: str,
        expr_config_path: str,
        task_config_path: str,
        platform_serializer_class: PlatformSerializer,
        save_dir: str = 'safe-autonomy-sims/safe_autonomy_sims/evaluation/animation/figs/',
        launch_dir_of_experiment: str = 'safe-autonomy-sims',
        tmp_dir: str = '/tmp/eval_results/animation_plots',
        start_index: int = 0
    ):

        self.checkpoint_path = checkpoint_path
        self.expr_config_path = expr_config_path
        self.task_config_path = task_config_path
        self.platform_serializer_class = platform_serializer_class
        self.save_dir = save_dir
        self.launch_dir_of_experiment = launch_dir_of_experiment
        self.tmp_dir = tmp_dir
        self.start_index = start_index

        os.makedirs(self.tmp_dir, exist_ok=True)
        os.makedirs(self.save_dir, exist_ok=True)
        self.filenames: list[str] = []
        self.filetype = 'png'

        self.metrics_config = {
            "agent": {
                "__default__": [
                    {
                        "name": "ObservationVector",
                        "functor": "safe_autonomy_sims.evaluation.general_metrics.ObservationVector",
                        "config": {
                            "description": "Observations"
                        }
                    },
                    {
                        "name": "ControlVector",
                        "functor": "safe_autonomy_sims.evaluation.general_metrics.ControlVector",
                        "config": {
                            "description": "Actions"
                        }
                    },
                ]
            }
        }

        self.default_test_case_manager_config = {
            "type": "corl.evaluation.runners.section_factories.test_cases.default_strategy.DefaultStrategy",
            "config": {
                "num_test_cases": 1
            }
        }

    def make_animation(
        self,
        dataframe: typing.Union[pd.DataFrame, None] = None,
        filetype: str = 'png',
    ):
        """
        Run one evaluation episode using the checkpoint and make an animation.

        Parameters
        ----------
        dataframe: typing.Union[pd.DataFrame, None]
            A pre-existing dataframe to be used, to bypass running and evaluation episode
        filetype: str
            Type of file to save the animation/plot. Use 'mp4' or 'gif' for animations, and 'png' for plots
        """
        self.filetype = filetype
        assert self.filetype in ['png', 'gif', 'mp4'], "filetype must be 'png', 'gif', or 'mp4'"

        if dataframe is None:
            dataframe = run_one_evaluation(
                self.task_config_path,
                self.expr_config_path,
                self.launch_dir_of_experiment,
                self.metrics_config,
                self.checkpoint_path,
                self.platform_serializer_class,
                test_case_manager_config=self.default_test_case_manager_config
            )
            pd.to_pickle(dataframe, os.path.join(self.save_dir, 'episode_dataframe.pkl'))

        data = self.get_data_from_dataframe(dataframe)

        axes = self.setup_plots(data)

        self.filenames = []
        print('Animation Progress:')
        for i in range(self.start_index, len(dataframe['ObservationVector'].values[0])):
            self.make_plots_in_loop(axes, data, i)

            # Print progress
            var = (i + 1) / len(dataframe['ObservationVector'].values[0])
            sys.stdout.write('\r')
            sys.stdout.write(f"[{'=' * int(var * 20) : <20}] {var * 100}%")
            sys.stdout.flush()

            if self.filetype != 'png':
                # Save figure
                filename = os.path.join(self.tmp_dir, 'frame' + str(i) + '.png')
                self.filenames.append(filename)
                plt.tight_layout()
                plt.savefig(filename)

                # Remove temporary elements
                self.remove_temp_elements(axes)

        print('\n')
        self.print_metrics(data)

        if self.filetype == 'png':
            # Make plot
            plt.tight_layout()
            plt.savefig(os.path.join(self.save_dir, 'episode_plot.png'))
        else:
            # Make video
            images: list[ArrayLike] = [imageio.imread(filename) for filename in self.filenames]
            imageio.mimwrite(os.path.join(self.save_dir, 'episode_animation.' + self.filetype), ims=images, fps=30)

            # Remove temp files
            for filename in set(self.filenames):
                os.remove(filename)
            os.rmdir(self.tmp_dir)

    def get_data_from_dataframe(self, dataframe: pd.DataFrame) -> typing.Any:
        """
        Get data from the dataframe into another form

        Parameters
        ----------
        dataframe: pd.DataFrame
            Dataframe created by running an evaluation episode

        Returns
        -------
        typing.Any
            Episode data
        """
        return dataframe

    @abc.abstractmethod
    def setup_plots(self, data: typing.Any) -> typing.Any:
        """
        Setup and return all axes

        Parameters
        ----------
        data: typing.Any
            Episode data

        Returns
        -------
        typing.Any
            Axes
        """
        raise NotImplementedError

    @abc.abstractmethod
    def make_plots_in_loop(self, axes: typing.Any, data: typing.Any, i: int):
        """
        Add data to the plots inside the loop.

        Parameters
        ----------
        axes: typing.Any
            Axes
        data: typing.Any
            Episode data
        i: int
            Current loop iteration
        """
        raise NotImplementedError

    def remove_temp_elements(self, axes: typing.Any):
        """
        Remove any temporary animation elements

        Parameters
        ----------
        axes: typing.Any
            Axes
        """

    def print_metrics(self, data: typing.Any):
        """
        Print any final metrics for the episode

        Parameters
        ----------
        data: typing.Any
            Episode data
        """



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core  Extension.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

This package implements CoRL glues for the docking and inspection
environments.

Glues connect an agent [platform](../platforms/index.md) to an RL
training framework by providing an endpoint for observations and
actions.

A glue may transform and send data from a platform part to an RL
training framework as an observation. The full set of glues which
provide observations over all agents in the environment defines the
environment's observation space.

A glue may also transform and send data from a training framework to
a platform part as an action. The full set of glues which provide
actions over all agents in the environment defines the environment's
action space.

Multiple glues may be used in a decorator pattern to reuse common
transformation logic.
"""
